# coding: utf-8
from atropos.util.colorspace import encode, decode
from atropos.commands import execute_cli
from .utils import run, datapath

# If there are any unknown characters in the test sequence,
# round tripping will only work if all characters after the
# first unknown character are also unknown:
# encode("TNGN") == "T444", but
# decode("T444") == "TNNN".

sequences = [
    "",
    "C",
    "ACGGTC",
    "TN",
    "TN.",
    "TNN.N",
    "CCGGCAGCATTCATTACGACAACGTGGCACCGTGTTTTCTCGGTGGTA",
    "TGCAGTTGATGATCGAAGAAAACGACATCATCAGCCAGCAAGTGC",
    "CAGGGTTTGATGAGTGGCTGTGGGTGCTGGCGTATCCGGG"
    ]


def test_encode():
    assert encode("AA") == "A0"
    assert encode("AC") == "A1"
    assert encode("AG") == "A2"
    assert encode("AT") == "A3"
    assert encode("CA") == "C1"
    assert encode("CC") == "C0"
    assert encode("CG") == "C3"
    assert encode("CT") == "C2"
    assert encode("GA") == "G2"
    assert encode("GC") == "G3"
    assert encode("GG") == "G0"
    assert encode("GT") == "G1"
    assert encode("TA") == "T3"
    assert encode("TC") == "T2"
    assert encode("TG") == "T1"
    assert encode("TT") == "T0"

    assert encode("TN") == "T4"
    assert encode("NT") == "N4"
    assert encode("NN") == "N4"

    assert encode("ACGGTC") == "A13012"
    assert encode("TTT.N") == "T0044"
    assert encode("TTNT.N") == "T04444"


def test_decode():
    for s in sequences:
        expected = s.replace('.', 'N')
        encoded = encode(s)
        assert decode(encoded) == expected
    assert decode('A.') == 'AN'
    assert decode('C.') == 'CN'
    assert decode('G.') == 'GN'
    assert decode('T.') == 'TN'


def test_qualtrim_csfastaqual():
    '''-q with csfasta/qual files'''
    run("-c -q 10", "solidqual.fastq", "solid.csfasta", qualfile='solid.qual')


def test_E3M():
    '''Read the E3M dataset'''
    # not really colorspace, but a fasta/qual file pair
    execute_cli(['-o', '/dev/null', '-se', datapath("E3M.fasta"), '-sq', datapath("E3M.qual")])


def test_bwa():
    '''MAQ-/BWA-compatible output'''
    run("-c -e 0.12 -a 330201030313112312 -x 552: --maq", "solidmaq.fastq", "solid.csfasta", qualfile='solid.qual')


def test_bfast():
    '''BFAST-compatible output'''
    run("-c -e 0.12 -a 330201030313112312 -x abc: --strip-f3", "solidbfast.fastq", "solid.csfasta", qualfile='solid.qual')


def test_trim_095():
    '''some reads properly trimmed since atropos 0.9.5'''
    run("-c -e 0.122 -a 330201030313112312", "solid.fasta", "solid.fasta")


def test_solid():
    run("-c -e 0.122 -a 330201030313112312", "solid.fastq", "solid.fastq")


def test_solid_basespace_adapter():
    '''colorspace adapter given in basespace'''
    run("-c -e 0.122 -a CGCCTTGGCCGTACAGCAG", "solid.fastq", "solid.fastq")


def test_solid5p():
    '''test 5' colorspace adapter'''
    # this is not a real adapter, just a random string
    # in colorspace: C0302201212322332333
    run("-c -e 0.1 --trim-primer -g CCGGAGGTCAGCTCGCTATA", "solid5p.fasta", "solid5p.fasta")


def test_solid5p_prefix_notrim():
    '''test anchored 5' colorspace adapter, no primer trimming'''
    run("-c -e 0.1 -g ^CCGGAGGTCAGCTCGCTATA", "solid5p-anchored.notrim.fasta", "solid5p.fasta")


def test_solid5p_prefix():
    '''test anchored 5' colorspace adapter'''
    run("-c -e 0.1 --trim-primer -g ^CCGGAGGTCAGCTCGCTATA", "solid5p-anchored.fasta", "solid5p.fasta")


def test_solid5p_fastq():
    '''test 5' colorspace adapter'''
    # this is not a real adapter, just a random string
    # in colorspace: C0302201212322332333
    run("-c -e 0.1 --trim-primer -g CCGGAGGTCAGCTCGCTATA", "solid5p.fastq", "solid5p.fastq")


def test_solid5p_prefix_notrim_fastq():
    '''test anchored 5' colorspace adapter, no primer trimming'''
    run("-c -e 0.1 -g ^CCGGAGGTCAGCTCGCTATA", "solid5p-anchored.notrim.fastq", "solid5p.fastq")


def test_solid5p_prefix_fastq():
    '''test anchored 5' colorspace adapter'''
    run("-c -e 0.1 --trim-primer -g ^CCGGAGGTCAGCTCGCTATA", "solid5p-anchored.fastq", "solid5p.fastq")


def test_sra_fastq():
    '''test SRA-formatted colorspace FASTQ'''
    run("-c -e 0.1 --format sra-fastq -a CGCCTTGGCCGTACAGCAG", "sra.fastq", "sra.fastq")


def test_no_zero_cap():
    run("--no-zero-cap -c -e 0.122 -a CGCCTTGGCCGTACAGCAG", "solid-no-zerocap.fastq", "solid.fastq")



"""Report generator for trim command.
"""
import os
from atropos.commands.legacy_report import LegacyReportGenerator

class ReportGenerator(LegacyReportGenerator):
    template_path = os.path.join(os.path.dirname(__file__), 'templates')



# coding: utf-8
"""Routines for printing a text report. This is the legacy code for generating
text-based reports. This will eventually be deprecated in favor of the Jinja2
and MultiQC reports.
"""
import math
import textwrap
from atropos.util import truncate_string, weighted_median
from .reports import BaseReportGenerator

INDENT = '  '
PARAGRAPH = textwrap.TextWrapper()
INDENTED = textwrap.TextWrapper(initial_indent=INDENT, subsequent_indent=INDENT)

class Printer(object):
    """Manages printing to a file.
    
    Args:
        outfile: The output file.
        kwargs: Additional keyword arguments passed to the print function.
    """
    def __init__(self, outfile, indent=None, **kwargs):
        self.outfile = outfile
        self.indent = indent
        self.print_args = kwargs
    
    def __call__(self, *args, indent=None, **kwargs):
        indent = indent or self.indent
        if indent:
            self._print(indent, end='')
        self._print(*args, **kwargs)
    
    def _print(self, *args, **kwargs):
        if self.print_args:
            print_args = self.print_args.copy()
            print_args.update(kwargs)
        else:
            print_args = kwargs
        print(*args, file=self.outfile, **print_args)
    
    def newline(self):
        """Print a newline.
        """
        print(file=self.outfile)

class TitlePrinter(Printer):
    """Printer that formats titles.
    
    Args:
        outfile: The output file.
        levels: The formatting associated with different header levels.
        kwargs: Additional keyword arguments passed to the print function.
    """
    def __init__(
            self, outfile,
            levels=(('=', '='), ('-', '-'), ('-', None), ('~', None)),
            **kwargs):
        super().__init__(outfile, **kwargs)
        self.levels = levels
    
    def __call__(self, *title, level=None, newline=True, **kwargs):
        title = ' '.join(title)
        
        if level is not None:
            if level >= len(self.levels):
                raise ValueError("Invalid level: {}".format(level))
            underline, overline = self.levels[level]
            if overline is True:
                overline = underline
            width = len(title)
            if overline:
                self._print(overline * width, **kwargs)
        
        self._print(title, **kwargs)
        
        if level is not None and underline:
            self._print(underline * width, **kwargs)
        
        if newline:
            self.newline()

class RowPrinter(Printer):
    """Priter that formats rows in a table.
    
    Args:
        outfile: The output file.
        colwidths: Column widths.
        justification: Column justifiations ('<' for left, '>' for right).
        indent: Column indents.
        pct: Whether floats should be formatted as percentages.
        default: Default value for None's.
        kwargs: Additional keyword arguments passed to the print function.
    
    colwidths, justification, and indent can be longer or shorter than the
    number of arguments; if shorter, the last value in the list is repeated;
    if longer, the list is truncated.
    """
    def __init__(
            self, outfile, colwidths=10, justification=('<', '>'), indent='',
            pct=False, default=0, **kwargs):
        super().__init__(outfile, **kwargs)
        self.colwidths, self.justification, self.indent = (
            (arg,) if isinstance(arg, typ) else tuple(arg)
            for arg, typ in zip(
                (colwidths, justification, indent),
                (int, str, str)))
        self.pct = pct
        self.default = default
    
    def print_rows(self, *rows, header=None, **kwargs):
        """Print multiple rows. Automatically figures out column widths.
        
        Args:
            rows: Rows to print.
            header: Header row.
            kwargs: Additional keyword arguments to self.__call__.
        """
        colwidths = tuple(sizeof(*x) for x in zip(*rows))
        if header:
            header_widths = (sizeof(h) for h in header)
            colwidths = tuple(
                max(h, c)
                for h, c in zip(header_widths, colwidths))
            self(*header, colwidths=colwidths, header=True, **kwargs)
        for row in rows:
            self(*row, colwidths=colwidths)
    
    def __call__(
            self, *args, colwidths=None, extra_width=None, justification=None,
            extra_justification=None, indent=None, extra_indent=None,
            header=False, underline='-', pct=None, default=None, **kwargs):
        """Print a row.
        
        Args:
            args: Fields in the row.
            colwidths, justification, indent: Row-specific colwidths,
                justification, indent.
            extra_width, extra_justification, extra_indent: colwidth/
                justification/indent to use for extra fields.
            header: Whether this is a header row.
            underline: Whether to use an underline after the header row. Either
                a bool or a character.
            pct: Whether floating point values should be formatted as
                percentages.
            default: Default value.
            kwargs: Additional keyword arguments to pass to print.
        """
        ncols = len(args)
        if ncols == 0:
            self.newline()
            return
        
        if pct is None:
            pct = self.pct
        
        def adjust(arr, extra=None):
            """Adjust an array. If longer than the number of columns,
            truncate; if shorter, fill in by repeating the last element.
            """
            alen = len(arr)
            if alen == ncols:
                return arr
            elif alen > ncols:
                return arr[:ncols]
            else:
                return arr + ((extra or arr[-1],) * (ncols - alen))
        
        colwidths, justification, indent = (
            adjust(arr, extra)
            for arr, extra in zip(
                (
                    colwidths or self.colwidths,
                    justification or self.justification,
                    indent or self.indent),
                (extra_width, extra_justification, extra_indent)))
        
        # adjust colwidths if this is a header
        if header:
            colwidths = tuple(
                max(w, len(str(a)))
                for w, a in zip(colwidths, args))
        
        fmt_str = []
        fmt_args = []
        for i, (value, width, just, ind) in enumerate(
                zip(args, colwidths, justification, indent)):
            if value is None:
                value = default or self.default
            if isinstance(value, str):
                typ = 's'
                if len(value) > width:
                    value = truncate_string(value, width)
            elif isinstance(value, float):
                typ = '.1' + ('%' if pct else 'f')
            else:
                typ = ',d'
            fmt_str.append(
                ind + '{' + str(i) + ':' + just + str(width-len(ind)) +
                typ + '}')
            fmt_args.append(value)
        
        fmt_str = ' '.join(fmt_str)
        self._print(fmt_str.format(*fmt_args), **kwargs)
        
        if header:
            sepline = ' '.join((underline * width) for width in colwidths)
            self._print(sepline, **kwargs)

class LegacyReportGenerator(BaseReportGenerator):
    def generate_text_report(self, fmt, summary, outfile, **kwargs):
        if fmt == 'txt':
            generate_report(summary, outfile)
        else:
            super().generate_from_template(fmt, summary, outfile, **kwargs)

def generate_report(summary, outfile):
    """Generate a report.
    
    Args:
        summary: The summary dict.
        outfile: The output file name/object.
    """
    is_path = isinstance(outfile, str)
    if is_path:
        outfile = open(outfile, 'wt')
    try:
        print_summary_report(summary, outfile)
        if 'trim' in summary:
            print_trim_report(summary, outfile)
        if 'pre' in summary:
            print_pre_trim_report(summary, outfile)
        if 'post' in summary:
            print_post_trim_report(summary, outfile)
    finally:
        if is_path:
            outfile.close()

def print_summary_report(summary, outfile):
    """Print the top-level summary report.
    
    Args:
        summary: The summary dict.
        outfile: The output file object.
    """
    _print_title = TitlePrinter(outfile)
    _print = Printer(outfile)
    
    _print_title("Atropos", level=0)
    _print("Atropos version: {}".format(summary['version']))
    _print("Python version: {}".format(summary['python']))
    _print("Command line parameters: {} {}".format(
        summary['command'], " ".join(summary['options']['orig_args'])))
    _print()
    
    _print("Sample ID: {}".format(summary['sample_id']))
    _print("Input format: {}".format(summary['derived']['input_format']))
    _print("Input files:")
    for infile in summary['input']['input_names']:
        _print(infile, indent=INDENT)
    _print()
    
    timing = summary['timing']
    total = summary["total_record_count"]
    wctime = ["Wallclock time: {:.2F} s".format(timing["wallclock"])]
    if total > 0:
        wctime.append("({0:.0F} us/read; {1:.2F} M reads/minute)".format(
            1E6 * timing["wallclock"] / total,
            total / timing["wallclock"] * 60 / 1E6))
    _print("Start time: {}".format(timing['start']))
    _print(*wctime)
    _print("CPU time (main process): {0:.2F} s".format(timing['cpu']))
    _print()

def print_trim_report(summary, outfile):
    """Print the trimming report.
    
    Args:
        summary: Summary dict.
        outfile: Open output stream.
    """
    paired = summary["options"]["paired"]
    pairs_or_reads = "Pairs" if paired else "Reads"
    total_bp = sum(summary['total_bp_counts'])
    max_width = len(str(total_bp))
    # account for commas
    max_width += (max_width // 3)
    
    _print_title = TitlePrinter(outfile)
    _print = RowPrinter(outfile, (35, max_width))
    
    total = summary["total_record_count"]
    if total == 0:
        _print(
            "No reads processed! Either your input file is empty or you "
            "used the wrong -f/--format parameter.")
        return
    
    modifiers, filters, formatters = (
        summary['trim'][key]
        for key in ('modifiers', 'filters', 'formatters'))
    adapter_cutter = None
    for modifier_dict in modifiers.values():
        if 'adapters' in modifier_dict:
            adapter_cutter = modifier_dict
            break
    correction_enabled = summary["options"]["correct_mismatches"]
    corrected = None
    trimmers = []
    for name, mod in modifiers.items():
        if 'bp_trimmed' in mod:
            trimmers.append((name, mod))
        if correction_enabled and 'records_corrected' in mod:
            corrected = mod
    
    _print_title("Trimming", level=1)
    _print(pairs_or_reads, 'records', 'fraction', header=True)
    _print(
        "Total {} processed".format('read pairs' if paired else 'reads'),
        total)
    if adapter_cutter:
        if paired:
            for read in range(2):
                _print(
                    "Read {} with adapter".format(read+1),
                    adapter_cutter['records_with_adapters'][read],
                    adapter_cutter['fraction_records_with_adapters'][read],
                    indent=(INDENT, ''), pct=True)
        else:
            _print(
                "Reads with adapter",
                adapter_cutter['records_with_adapters'][0],
                adapter_cutter['fraction_records_with_adapters'][0],
                pct=True)
    
    def _print_filter(name, sep):
        if name in filters:
            _print(
                "{} {} {}".format(pairs_or_reads, sep, name.replace('_', ' ')),
                filters[name]['records_filtered'],
                filters[name]['fraction_records_filtered'], pct=True)
    
    _print_filter('too_short', 'that were')
    _print_filter('too_long', 'that were')
    _print_filter('too_many_n', 'with')
    
    _print(
        '{} written (passing filters)'.format(pairs_or_reads),
        formatters['records_written'], formatters['fraction_records_written'],
        pct=True)
    
    if corrected:
        _print(
            'Pairs corrected', corrected['records_corrected'],
            corrected['fraction_records_corrected'], pct=True)
    
    _print()
    _print("Base pairs", 'bp', 'fraction', header=True)
    
    _print("Total bp processed", total_bp)
    if paired:
        for read in range(2):
            _print(
                "Read {}".format(read+1), summary['total_bp_counts'][read],
                indent=(INDENT, ''))
    
    def _print_bp(title, data, key, default=0):
        if paired:
            _print(
                title, data['total_{}'.format(key)],
                data['fraction_total_{}'.format(key)], pct=True)
            for read in range(2):
                _print(
                    "Read {}".format(read+1),
                    data[key][read],
                    data['fraction_{}'.format(key)][read],
                    indent=(INDENT, ''), pct=True, default=default)
        else:
            _print(
                title, data[key][0], data['fraction_{}'.format(key)][0],
                pct=True, default=default)
    
    for name, mod in trimmers:
        _print_bp(mod['desc'], mod, 'bp_trimmed')
    
    _print_bp("Total bp written (filtered)", formatters, 'bp_written')
    
    if corrected:
        _print_bp("Total bp corrected", formatters, 'bp_corrected')
    
    if adapter_cutter:
        _print()
        adapters = adapter_cutter['adapters']
        print_adapter_report(adapters, outfile, paired, total, max_width)

def print_adapter_report(adapters, outfile, paired, total_records, max_width):
    """Print details for a adapters.
    
    Args:
        adapters: Sequence of adapter info dicts.
        outfile: The output file
        paired: Whether the data is paired-end.
        
        max_width: Max column width.
    """
    adapter_lenghts = []
    for pair in adapters:
        if pair:
            for adapter in pair.values():
                if adapter['where']['name'] == 'linked':
                    adapter_lenghts.append(3 + len(
                        adapter['front_sequence'] +
                        adapter['back_sequence']))
                else:
                    adapter_lenghts.append(len(adapter['sequence']))
    max_seq_len = max(adapter_lenghts)
    
    _print = Printer(outfile)
    _print_title = TitlePrinter(outfile)
    _print_adj = RowPrinter(outfile, (12, 5), pct=True, indent=(INDENT, ''))
    
    seq_printer = RowPrinter(
        outfile, (max_seq_len, 14, 3, max_width), ('<', '<', '>'))
    hist_printer = RowPrinter(outfile, justification=('>', '>', '>', '>', '<'))
    
    def print_error_ranges(adapter_length, error_rate):
        """Print max number of errors for ranges of adapter match lengths.
        """
        _print("No. of allowed errors:")
        prev = 0
        for errors in range(1, int(error_rate * adapter_length) + 1):
            range_start = int(errors / error_rate)
            _print(
                "{0}-{1} bp: {2};".format(prev, range_start - 1, errors - 1),
                end=' ')
            prev = range_start
        if prev == adapter_length:
            _print("{0} bp: {1}".format(
                adapter_length, int(error_rate * adapter_length)))
        else:
            _print("{0}-{1} bp: {2}".format(
                prev, adapter_length, int(error_rate * adapter_length)))
        _print()
    
    # TODO: compute max width for error counts column, print secondary header
    # with number of errors, space-pad each sub-column, think about whether to
    # show trailing zeroes.
    def print_histogram(data, adapter_length, num_reads, error_rate, errors, match_probabilities):
        """Print a histogram. Also, print the no. of reads expected to be
        trimmed by chance (assuming a uniform distribution of nucleotides in
        the reads).
        
        Args:
            data: A dictionary mapping lengths of trimmed sequences to their
                respective frequency
            adapter_length: adapter length
            num_reads: total no. of reads.
            error_rate: Max error rate.
            errors: Histogram of actual numbers of errors.
            match_probabilities: List of random match proabilities for each
                position in the adapter seqeunce.
        """
        hist = []
        for length, count in data.items():
            # when length surpasses adapter_length, the probability does not
            # increase anymore
            estimated = (
                num_reads * match_probabilities[min(length, adapter_length)])
            errs = ' '.join(str(err) for err in errors['rows'][length])
            hist.append((
                length, count, estimated,
                int(error_rate * min(length, adapter_length)), errs))
        
        hist_printer.print_rows(
            *hist,
            header=("length", "count", "expect", "max.err", "error counts"))
        hist_printer.newline()
    
    def print_adjacent_bases(bases):
        """Print a summary of the bases preceding removed adapter sequences.
        Print a warning if one of the bases is overrepresented and there are
        at least 20 preceding bases available.
        
        Return:
            True if a warning was printed.
        """
        total = sum(bases.values())
        if total == 0:
            return False
        _print('Bases preceding removed adapters:')
        warnbase = None
        for base in ['A', 'C', 'G', 'T', '']:
            base_label = base if base != '' else 'none/other'
            fraction = 1.0 * bases[base] / total
            _print_adj(base_label, fraction)
            if fraction > 0.8 and base != '':
                warnbase = base_label
        if total >= 20 and warnbase is not None:
            _print('WARNING:')
            _print(INDENTED.wrap(
                'The adapter is preceded by "{0}" extremely often. The '
                'provided adapter sequence may be incomplete. To fix the '
                'problem, add "{0}" to the beginning of the adapter '
                'sequence.'.format(warnbase)))
            _print()
            return True
        _print()
        return False
        
    warning = False
    for pair in range(2 if paired else 1):
        if adapters[pair] is None:
            continue
        
        header = "Adapter {}"
        if paired:
            header = ('First read: ' if pair == 0 else 'Second read: ') + header
        
        for name, adapter in adapters[pair].items():
            _print_title(header.format(name), level=1)
            
            where_name = adapter["where"]["name"]
            if where_name == "linked":
                front_len, back_len = [
                    len(adapter[s])
                    for s in ('front_sequence', 'back_sequence')]
                seq_printer.print_rows(
                    (
                        "{}...{}".format(
                            adapter["front_sequence"],
                            adapter["back_sequence"]),
                        "linked",
                        "{}+{}".format(front_len, back_len),
                        adapter["total_front"],
                        adapter["total_back"]),
                    header=(
                        "Sequence", "Type", "Length", "Trimmed (x)",
                        "Half matches (x)"))
            else:
                seq_len = len(adapter["sequence"])
                seq_printer.print_rows(
                    (
                        adapter["sequence"], adapter["where"]["desc"], seq_len,
                        adapter["total"]),
                    header=(
                        "Sequence", "Type", "Length", "Trimmed (x)"))
            
            _print()
            
            if adapter["total"] == 0:
                return
            
            if where_name == "anywhere":
                _print(
                    adapter["total_front"],
                    "times, it overlapped the 5' end of a read")
                _print(
                    adapter["total_back"],
                    "times, it overlapped the 3' end or was within the read")
                _print()
                print_error_ranges(seq_len, adapter["max_error_rate"])
                _print("Overview of removed sequences (5'):")
                print_histogram(
                    adapter["lengths_front"], seq_len, total_records,
                    adapter["max_error_rate"], adapter["errors_front"],
                    adapter['match_probabilities'])
                _print()
                _print("Overview of removed sequences (3' or within):")
                print_histogram(
                    adapter["lengths_back"], seq_len, total_records,
                    adapter["max_error_rate"], adapter["errors_back"],
                    adapter['match_probabilities'])
            
            elif where_name == "linked":
                print_error_ranges(front_len, adapter["front_max_error_rate"])
                print_error_ranges(back_len, adapter["back_max_error_rate"])
                _print("Overview of removed sequences at 5' end:")
                print_histogram(
                    adapter["front_lengths_front"], front_len, total_records,
                    adapter["front_max_error_rate"],
                    adapter["front_errors_front"], 
                    adapter['front_match_probabilities'])
                _print()
                _print("Overview of removed sequences at 3' end:")
                print_histogram(
                    adapter["back_lengths_back"], back_len, total_records,
                    adapter["back_max_error_rate"], adapter["back_errors_back"],
                    adapter['back_match_probabilities'])
            
            elif where_name in ("front", "prefix"):
                print_error_ranges(seq_len, adapter["max_error_rate"])
                _print("Overview of removed sequences:")
                print_histogram(
                    adapter["lengths_front"], seq_len, total_records,
                    adapter["max_error_rate"], adapter["errors_front"],
                    adapter['match_probabilities'])
            
            elif where_name in ("back", "suffix"):
                print_error_ranges(seq_len, adapter["max_error_rate"])
                warning = warning or print_adjacent_bases(
                    adapter["adjacent_bases"])
                _print("Overview of removed sequences:")
                print_histogram(
                    adapter["lengths_back"], seq_len, total_records,
                    adapter["max_error_rate"], adapter["errors_back"],
                    adapter['match_probabilities'])
    
    if warning:
        _print('WARNING:')
        _print(INDENTED.wrap(
            'One or more of your adapter sequences may be incomplete. '
            'Please see the detailed output above.'))

def print_pre_trim_report(summary, outfile):
    """Print pre-trimming stats.
    
    Args:
        summary: The summary dict.
        outfile: The output file.
    """
    pre = summary['pre']
    _print_title = TitlePrinter(outfile)
    _print = Printer(outfile)
    _print_title("Pre-trimming stats", level=1)
    for source, data in pre.items():
        _print_title("Source", level=3, newline=False)
        # TODO: When multi-file input is supported, this code will need to
        # get summary['input']['input_names'][source]
        for read, src in enumerate(summary['input']['input_names'], 1):
            if src is None:
                continue
            _print("Read {}: {}".format(read, src))
        _print()
        print_stats_report(data, outfile)
    
def print_post_trim_report(summary, outfile):
    """Print post-trimming stats.
    
    Args:
        summary: The summary dict.
        outfile: The output file.
    """
    post = summary['post']
    _print_title = TitlePrinter(outfile)
    _print = Printer(outfile)
    _print_title("Post-trimming stats", level=1)
    for dest, stats in post.items():
        _print_title("Destination: {}".format(dest), level=2)
        for source, data in stats.items():
            _print_title("Source", level=3, newline=False)
            # TODO: When multi-file input is supported, this code will need to
            # get summary['input']['input_names'][source]
            for read, src in enumerate(summary['input']['input_names'], 1):
                if src is None:
                    continue
                _print("Read {}: {}".format(read, src))
            _print()
            print_stats_report(data, outfile)

def print_stats_report(data, outfile):
    """Print stats.
    
    Args:
        data: The stats dict.
        outfile: The output file.
    """
    paired = 'read2' in data
    max_count = data['read1']['counts']
    if paired:
        max_count = max(max_count, data['read2']['counts'])
    max_width = len(str(max_count))
    # add space for commas and column separation
    max_width += (max_width // 3) + 1
    
    _print_title = TitlePrinter(outfile)
    _print = RowPrinter(outfile, (35, max_width))
    
    def _print_histogram(title, hist1, hist2=None):
        _print_title(title, level=2)
        if hist2:
            hist = (
                (key, hist1.get(key, 0), hist2.get(key, 0))
                for key in sorted(set(hist1.keys()) | set(hist2.keys())))
        else:
            hist = sorted(hist1.items(), key=lambda x: x[0])
        for histbin in hist:
            _print(*histbin)
    
    def _print_base_histogram(title, hist, extra_width=4, index_name='Pos'):
        _print_title(title, level=2)
        _print(
            index_name, *hist['columns'], header=True, extra_width=extra_width)
        for pos, row in hist['rows'].items():
            total_count = sum(row)
            base_pcts = (
                round(count * 100 / total_count, 1)
                for count in row)
            _print(pos, *base_pcts, extra_width=extra_width)
    
    def _print_tile_histogram(title, hist):
        ncol = len(hist['columns'])
        max_tile_width = max(
            4, len(str(math.ceil(data['read1']['counts'] / ncol)))) + 1
        _print_base_histogram(
            title, hist, extra_width=max_tile_width, index_name='Tile')
    
    def _print_tile_base_histogram(title, hist):
        """Print a histogram of position x tile, with values as the median
        base quality.
        """
        quals = hist['columns']
        tiles = hist['columns2']
        ncol = len(tiles)
        max_tile_width = max(
            4, len(str(math.ceil(data['read1']['counts'] / ncol)))) + 1
        _print_title(title, level=2)
        _print('Pos', *tiles, header=True, extra_width=max_tile_width)
        for pos, tiles in hist['rows'].items():
            # compute the weighted median for each tile at each position
            _print(
                pos,
                *(
                    weighted_median(quals, tile_counts)
                    for tile_counts in tiles.values()),
                extra_width=max_tile_width)
    
    _print('', 'Read1', 'Read2', header=True)
    
    # Sequence-level stats
    _print(
        "Read pairs:" if paired else "Reads:",
        data['read1']['counts'],
        data['read2']['counts'])
    _print()
    _print_histogram(
        "Sequence lengths:",
        data['read1']['lengths']['hist'],
        data['read2']['lengths']['hist'])
    _print()
    if 'qualities' in data['read1']:
        _print_histogram(
            "Sequence qualities:",
            data['read1']['qualities']['hist'],
            data['read2']['qualities']['hist'])
        _print()
    _print_histogram(
        "Sequence GC content (%)",
        data['read1']['gc']['hist'],
        data['read2']['gc']['hist'])
    _print()
    
    if 'tile_sequence_qualities' in data['read1']:
        _print_tile_histogram(
            "Read 1 per-tile sequence qualities (%)",
            data['read1']['tile_sequence_qualities'])
        _print()
        _print_tile_histogram(
            "Read 2 per-tile sequence qualities (%)",
            data['read2']['tile_sequence_qualities'])
        _print()
    
    # Base-level stats
    if 'base_qualities' in data['read1']:
        _print_base_histogram(
            "Read 1 base qualities (%)",
            data['read1']['base_qualities'])
        _print()
        _print_base_histogram(
            "Read 2 base qualities (%)",
            data['read2']['base_qualities'])
        _print()
    _print_base_histogram(
        "Read 1 base composition (%)",
        data['read1']['bases'])
    _print()
    _print_base_histogram(
        "Read 2 base composition (%)",
        data['read2']['bases'])
    _print()
    if 'tile_base_qualities' in data['read1']:
        _print_tile_base_histogram(
            "Read 1 per-tile base qualities (%)",
            data['read1']['tile_base_qualities'])
        _print()
        _print_tile_base_histogram(
            "Read 2 per-tile base qualities (%)",
            data['read2']['tile_base_qualities'])
        _print()

def sizeof(*x):
    """Returns the largest string size of all objects in x, where x is a
    sequence of string or numeric values.
    """
    if isinstance(x[0], str):
        return max(len(s) for s in x)
    elif isinstance(x[0], int):
        return len(str(max(x)))
    elif isinstance(x[0], float):
        return len(str(round(max(x), 1)))



"""Classes for formatting and writing trimmed reads to output.
"""
import sys
from atropos.io import STDOUT, xopen, open_output
from atropos.io.compression import splitext_compressed
from atropos.io.seqio import create_seq_formatter
from .filters import NoFilter

class Writers(object):
    """Manages writing to one or more outputs.
    
    Args:
        force_create: Whether empty output files should be created.
    """
    def __init__(self, force_create):
        self.writers = {}
        self.force_create = force_create
        self.suffix = None
    
    def get_writer(self, file_desc, compressed=False):
        """Create the writer for a file descriptor if it does not already
        exist.
        
        Args:
            file_desc: File descriptor. If `compressed==True`, this is a tuple
                (path, mode), otherwise it's only a path.
            compressed: Whether data has already been compressed.
        
        Returns:
            The writer.
        """
        if compressed:
            path, mode = file_desc
        else:
            path = file_desc
        
        if path not in self.writers:
            if self.suffix:
                real_path = add_suffix_to_path(path, self.suffix)
            else:
                real_path = path
            # TODO: test whether O_NONBLOCK allows non-blocking write to NFS
            if compressed:
                self.writers[path] = open_output(real_path, mode)
            else:
                self.writers[path] = xopen(real_path, "w")
        
        return self.writers[path]
    
    def write_result(self, result, compressed=False):
        """Write results to output.
        
        Args:
            result: Dict with keys being file descriptors and values being data
                (either bytes or strings). Strings are expected to already have
                appropriate line-endings.
            compressed: Whether data has already been compressed.
        """
        for file_desc, data in result.items():
            self.write(file_desc, data, compressed)
    
    def write(self, file_desc, data, compressed=False):
        """Write data to output. If the specified path has not been seen before,
        the output is opened.
        
        Args:
            file_desc: File descriptor. If `compressed==True`, this is a tuple
                (path, mode), otherwise it's only a path.
            data: The data to write.
            compressed: Whether data has already been compressed.
        """
        self.get_writer(file_desc, compressed).write(data)
    
    def close(self):
        """Close all outputs.
        """
        for path in self.force_create:
            if path not in self.writers and path != STDOUT:
                with open_output(path, "w"):
                    pass
        for writer in self.writers.values():
            if writer not in (sys.stdout, sys.stderr):
                writer.close()

class Formatters(object):
    """Manages multiple formatters.
    
    Args:
        output: The output file name template.
        seq_formatter_args: Additional arguments to pass to the formatter
            constructor.
    """
    def __init__(self, output, seq_formatter_args):
        self.output = output
        self.multiplexed = output is not None and '{name}' in output
        self.seq_formatter_args = seq_formatter_args
        self.seq_formatters = {}
        self.mux_formatters = {}
        self.info_formatters = []
        self.discarded = 0
    
    def add_seq_formatter(self, filter_type, file1, file2=None):
        """Add a formatter.
        
        Args:
            filter_type: The type of filter that triggers writing with the
                formatter.
            file1, file2: The output file(s).
        """
        self.seq_formatters[filter_type] = create_seq_formatter(
            file1, file2, **self.seq_formatter_args)
    
    def add_info_formatter(self, formatter):
        """Add a formatter for one of the delimited detail files
        (rest, info, wildcard).
        """
        self.info_formatters.append(formatter)
    
    def get_mux_formatter(self, name):
        """Returns the formatter associated with the given name (barcode) when
        running in multiplexed mode.
        """
        assert self.multiplexed
        if name not in self.mux_formatters:
            path = self.output.format(name=name)
            self.mux_formatters[name] = create_seq_formatter(
                path, **self.seq_formatter_args)
        return self.mux_formatters[name]
    
    def get_seq_formatters(self):
        """Returns a set containing all formatters that have handled at least
        one record.
        """
        return (
            set(f for f in self.seq_formatters.values() if f.written > 0) |
            set(f for f in self.mux_formatters.values() if f.written > 0))
    
    def format(self, result, dest, read1, read2=None):
        """Format read(s) and add to a result dict. Also writes info records
        to any registered info formatters.
        
        Args:
            result: The result dict.
            dest: The destination (filter type).
            read1, read2: The read(s).
        """
        if self.multiplexed and (dest == NoFilter) and read1.match:
            name = read1.match.adapter.name
            formatter = self.get_mux_formatter(name)
            formatter.format(result, read1, read2)
        elif dest in self.seq_formatters:
            self.seq_formatters[dest].format(result, read1, read2)
        else:
            self.discarded += 1
        
        for fmtr in self.info_formatters:
            fmtr.format(result, read1)
            if read2:
                fmtr.format(result, read2)
    
    def summarize(self):
        """Returns a summary dict.
        """
        seq_formatters = self.get_seq_formatters()
        return dict(
            records_written=sum(f.written for f in seq_formatters),
            bp_written=[
                sum(f.read1_bp for f in seq_formatters),
                sum(f.read2_bp for f in seq_formatters)
            ])

class DelimFormatter(object):
    """Base class for formatters that write to a delimited file.
    
    Args:
        path: The output file path.
        delim: The field delimiter.
    """
    def __init__(self, path, delim=' '):
        self.path = path
        self.delim = delim
    
    def format(self, result, read):
        """Format a read and add it to `result`.
        """
        raise NotImplementedError()
    
    def _format(self, result, fields):
        result[self.path].append("".join((
            self.delim.join(str(f) for f in fields),
            "\n")))

class RestFormatter(DelimFormatter):
    """Rest file formatter.
    """
    def format(self, result, read):
        if read.match:
            rest = read.match.rest()
            if len(rest) > 0:
                self._format(result, (rest, read.name))

class InfoFormatter(DelimFormatter):
    """Info file formatter.
    """
    def __init__(self, path):
        super(InfoFormatter, self).__init__(path, delim='\t')
    
    def format(self, result, read):
        if read.match:
            for match_info in read.match_info:
                self._format(result, match_info[0:11])
        else:
            seq = read.sequence
            qualities = read.qualities if read.qualities is not None else ''
            self._format(result, (read.name, -1, seq, qualities))

class WildcardFormatter(DelimFormatter):
    """Wildcard file formatter.
    """
    def format(self, result, read):
        if read.match:
            self._format(result, (read.match.wildcards(), read.name))

def add_suffix_to_path(path, suffix):
    """
    Add the suffix (str or int) after the file name but
    before the extension.
    """
    name, ext1, ext2 = splitext_compressed(path)
    return "{}{}{}{}".format(name, suffix, ext1, ext2 or "")



"""Classes and methods to support parallelization of operations.
"""
import inspect
import logging
from multiprocessing import Process, Value, Queue
import os
from queue import Empty, Full
import time
from atropos import AtroposError
from atropos.util import run_interruptible

RETRY_INTERVAL = 5
"""Max time to wait between retrying operations."""

# Control values
CONTROL_ACTIVE = 0
"""Controlled process should run normally."""
CONTROL_ERROR = -1
"""Controlled process should exit."""

class MulticoreError(AtroposError):
    """Base error for parallel processes.
    """
    pass

class Control(object):
    """Shared (long) value for passing control information between main and
    worker threads.
    
    Args:
        initial_value: Initial value of the shared control variable.
    """
    def __init__(self, initial_value=CONTROL_ACTIVE):
        self.control = Value('l', initial_value)
    
    def check_value(self, value, lock=False):
        """Check that the current control value == `value`.
        
        Args:
            value: The value to check.
            lock: Whether to lock the shared variable before checking.
        
        Returns:
            True if the values are equal.
        """
        return self.get_value(lock=lock) == value
    
    def check_value_positive(self, lock=False):
        """Check that the current control value is positive.
        
        Args:
            lock: Whether to lock the shared variable before checking.
        """
        return self.get_value(lock=lock) > 0
    
    def get_value(self, lock=True):
        """Returns the current control value.
        
        Args:
            lock: Whether to lock the shared variable before checking.
        """
        if lock:
            with self.control.get_lock():
                return self.control.value
        else:
            return self.control.value
    
    def set_value(self, value):
        """Set the control value. The shared variable is always locked.
        
        Args:
            value: The value to set.
        """
        with self.control.get_lock():
            self.control.value = value

class PendingQueue(object):
    """Queue for items with sequentially increasing priority. An item whose
    priority is below the current level is queued. Pop returns the item with
    the current priority and increments the current priority.
    
    Args:
        max_size: Maximum queue size; None == infinite.
    """
    def __init__(self, max_size=None):
        self.queue = {}
        self.max_size = max_size
        self.min_priority = None
    
    def push(self, priority, value):
        """Add an item to the queue with priority.
        
        Args:
            priority: An integer that determines the placement of `value` in
                the queue. Must be unique.
            value: The value to queue.
        
        Raises:
            Full if queue is full.
        """
        if self.full:
            raise Full()
        if priority in self.queue:
            raise ValueError("Duplicate priority value: {}".format(priority))
        self.queue[priority] = value
        if self.min_priority is None or priority < self.min_priority:
            self.min_priority = priority
    
    def pop(self):
        """Remove and return the item in the queue with lowest priority.
        
        Raises:
            Empty if the queue is emtpy.
        """
        if self.empty:
            raise Empty()
        value = self.queue.pop(self.min_priority)
        if self.empty:
            self.min_priority = None
        else:
            self.min_priority = min(self.queue.keys())
        return value
    
    @property
    def full(self):
        """Whether the queue is full.
        """
        return self.max_size and len(self.queue) >= self.max_size
    
    @property
    def empty(self):
        """Whether the queue is empty.
        """
        return len(self.queue) == 0

class ParallelPipelineMixin(object):
    """Mixin that implements the `start`, `finish`, and `process_batch` methods
    of :class:`Pipeline`.
    """
    def start(self, **kwargs):
        super().start(**kwargs)
        self.seen_batches = set()
    
    def process_batch(self, batch):
        self.seen_batches.add(batch[0]['index'])
        super().process_batch(batch)
    
    def finish(self, summary, worker=None):
        super().finish(summary, worker=worker)
        logging.getLogger().debug(
            "%s finished; processed %d batches, %d reads",
            worker.name, len(self.seen_batches),
            sum(self.record_counts.values()))

class WorkerProcess(Process):
    """Parent class for worker processes that execute Pipelines.
    
    Args:
        index: A unique ID for the process.
        input_queue: Queue with batches of records to process.
        pipeline: The pipeline to execute.
        summary_queue: Queue where summary information is written.
        timeout: Time to wait upon queue full/empty.
    """
    def __init__(self, index, input_queue, pipeline, summary_queue, timeout):
        super().__init__(name="Worker process {}".format(index))
        self.index = index
        self.input_queue = input_queue
        self.pipeline = pipeline
        self.summary_queue = summary_queue
        self.timeout = timeout
    
    def run(self):
        logging.getLogger().debug(
            "%s running under pid %d", self.name, os.getpid())
        
        summary = {}
        
        def iter_batches():
            """Deque and yield batches.
            """
            while True:
                batch = dequeue(
                    self.input_queue,
                    wait_message="{} waiting on batch {{}}".format(self.name),
                    timeout=self.timeout)
                yield batch
        
        def enqueue_summary():
            """Enqueue a summary dict.
            """
            enqueue(
                self.summary_queue,
                (self.index, self.pipeline.seen_batches, summary),
                wait_message="{} waiting to queue summary {{}}".format(
                    self.name),
                timeout=self.timeout
            )
        
        try:
            self.pipeline.start(worker=self)
            
            try:
                for batch in iter_batches():
                    if batch is None:
                        break
                    logging.getLogger().debug(
                        "%s processing batch of size %d",
                        self.name, batch[0]['size'])
                    self.pipeline.process_batch(batch)
            finally:
                self.pipeline.finish(summary, worker=self)
            
            logging.getLogger().debug("%s finished normally", self.name)
        except Exception as err:
            logging.getLogger().error(
                "Unexpected error in %s", self.name, exc_info=True)
            summary['exception'] = err
        
        logging.getLogger().debug("%s sending summary", self.name)
        enqueue_summary()

class ParallelPipelineRunner(object):
    """Run a pipeline in parallel.
    
    Args:
        reader: A :class:`BatchReader`.
        pipeline: A :class:`Pipeline`.
        threads: Number of threads to use. If None, the value will be taken
            from command_runner.
    """
    def __init__(self, command_runner, pipeline, threads=None):
        self.command_runner = command_runner
        self.pipeline = pipeline
        self.threads = threads or command_runner.threads
        self.timeout = max(command_runner.process_timeout, RETRY_INTERVAL)
        # Queue by which batches of reads are sent to worker processes
        self.input_queue = Queue(command_runner.read_queue_size)
        # Queue for processes to send summary information back to main process
        self.summary_queue = Queue(self.threads)
        self.worker_processes = None
        self.num_batches = None
        self.seen_summaries = None
        self.seen_batches = None
    
    def ensure_alive(self):
        """Callback when enqueue times out.
        """
        ensure_processes(self.worker_processes)
    
    def after_enqueue(self):
        """Called after all batches are queued.
        """
        pass
    
    def finish(self):
        """Called at the end of the run.
        """
        pass
    
    def run(self):
        """Run the pipeline.
        
        Returns:
            The return code.
        """
        retcode = run_interruptible(self)
        self.terminate(retcode)
        return retcode
    
    def terminate(self, retcode):
        """Terminates all processes.
        
        Args:
            retcode: The return code.
        """
        # notify all threads that they should stop
        logging.getLogger().debug("Exiting all processes")
        for process in self.worker_processes:
            kill(process, retcode, self.timeout)
    
    def __call__(self):
        # Start worker processes, reserve a thread for the reader process,
        # which we will get back after it completes
        worker_args = (
            self.input_queue, self.pipeline, self.summary_queue, self.timeout)
        self.worker_processes = launch_workers(self.threads - 1, worker_args)
        
        self.num_batches = enqueue_all(
            self.command_runner, self.input_queue, self.timeout,
            self.ensure_alive)
        
        logging.getLogger().debug(
            "Main loop complete; saw %d batches", self.num_batches)
        
        # Tell the worker processes no more input is coming
        enqueue_all(
            (None,) * self.threads, self.input_queue, self.timeout,
            self.ensure_alive)
        
        self.after_enqueue()
        
        # Now that the reader process is done, it essentially
        # frees up another thread to use for a worker
        self.worker_processes.extend(
            launch_workers(1, worker_args, offset=self.threads-1))
        
        # Wait for all summaries to be available on queue
        def summary_timeout_callback():
            """Ensure that workers are still alive.
            """
            try:
                ensure_processes(
                    self.worker_processes,
                    "Workers are still alive and haven't returned summaries: {}",
                    alive=False)
            except Exception as err:
                logging.getLogger().error(err)
            
        wait_on(
            self.summary_queue.full,
            wait_message="Waiting on worker summaries {}",
            timeout=self.timeout,
            wait=True,
            timeout_callback=summary_timeout_callback)
        
        # Process summary information from worker processes
        logging.getLogger().debug(
            "Processing summary information from worker processes")
        
        self.seen_summaries = set()
        self.seen_batches = set()
        
        def summary_fail_callback():
            """Raises AtroposError with workers that did not report summaries.
            """
            missing_summaries = (
                set(range(1, self.threads)) - self.seen_summaries)
            raise AtroposError(
                "Missing summaries from processes %s",
                ",".join(str(summ) for summ in missing_summaries))
        
        for _ in range(1, self.threads+1):
            batch = dequeue(
                self.summary_queue, fail_callback=summary_fail_callback)
            worker_index, worker_batches, worker_summary = batch
            if worker_summary is None:
                raise MulticoreError(
                    "Worker process {} died unexpectedly".format(worker_index))
            elif (
                    'exception' in worker_summary and
                    worker_summary['exception'] is not None):
                raise AtroposError(
                    "Worker process {} died unexpectedly".format(worker_index),
                    worker_summary['exception'])
            else:
                logging.getLogger().debug(
                    "Processing summary for worker %d", worker_index)
            self.seen_summaries.add(worker_index)
            self.seen_batches |= worker_batches
            self.command_runner.summary.merge(worker_summary)
        
        # Check if any batches were missed
        if self.num_batches > 0:
            missing_batches = (
                set(range(1, self.num_batches+1)) - self.seen_batches)
            if len(missing_batches) > 0:
                raise AtroposError(
                    "Workers did not process batches {}".format(
                        ",".join(str(batch) for batch in missing_batches)))
        
        self.finish()

def launch_workers(num_workers, args=(), offset=0, worker_class=WorkerProcess):
    """Launch `n` workers. Each worker is initialized with an incremental
    index starting with `offset`, followed by `args`.
    """
    logging.getLogger().info("Starting %d worker processes", num_workers)
    # create workers
    workers = [worker_class(i+offset, *args) for i in range(num_workers)]
    # start workers
    for worker in workers:
        worker.start()
    return workers

def ensure_processes(
        processes, message="One or more process exited: {}", alive=True):
    """Raise an exception if all processes do not have the expected status
    (alive or dead).
    """
    is_alive = [worker.is_alive() for worker in processes]
    if alive != all(is_alive):
        raise MulticoreError(message.format(",".join(
            str(i) for i in range(len(is_alive)) if not is_alive[i])))

def wait_on(
        condition, *args, wait_message="Waiting {}", timeout=None,
        fail_callback=None, wait=None, timeout_callback=None):
    """Wait on a condition to be non-False.
    
    Args:
        condition: Function that returns either False or a non-False value.
        args: Args with which to call `condition`.
        wait_message: The message to log while `condition` is False.
        timeout: Number of seconds after which the log messages escalate from
            DEBUG to ERROR.
        fail_callback: Function that is called each time `condition` returns
            False.
        wait: Either a boolean or a wait function to execute. If True,
            `time.sleep(RETRY_INTERVAL)` is used as the wait function.
        timeout_callback: Eather a function that is called each time, or an
            Exception class that is raised the first time `condition`
            returns False after waiting for `timeout` seconds.
    """
    if wait is True:
        wait = lambda: time.sleep(RETRY_INTERVAL)
    elif isinstance(wait, int):
        wait_time = wait
        wait = lambda: time.sleep(wait_time)
    wait_start = None
    while True:
        result = condition(*args)
        if result is not False:
            return result
        if fail_callback:
            fail_callback()
        now = time.time()
        if not wait_start:
            wait_start = now
        else:
            waiting = now - wait_start
            msg = wait_message.format(
                "for {} seconds".format(round(waiting, 1)))
            if timeout is not None and waiting >= timeout:
                logging.getLogger().error(msg)
                if timeout_callback:
                    if inspect.isclass(timeout_callback):
                        raise timeout_callback()
                    else:
                        timeout_callback()
            else:
                logging.getLogger().debug(msg)
            if wait:
                wait()

def wait_on_process(process, timeout, terminate=False):
    """Wait on a process to terminate.
    
    Args:
        process: The process on which to wait.
        timeout: Number of seconds to wait for process to terminate.
        terminate: Whether to force the process to terminate after `timeout`
            seconds.
    """
    timeout_callback = lambda: process.terminate() if terminate else None
    return wait_on(
        lambda: not process.is_alive(),
        wait_message="Waiting on {} to terminate {{}}".format(process.name),
        timeout=timeout,
        wait=lambda: process.join(RETRY_INTERVAL),
        timeout_callback=timeout_callback)

def enqueue(
        queue, item, wait_message="Waiting to enqueue item {}",
        block_timeout=RETRY_INTERVAL, **kwargs):
    """Enqueue an item, using `wait_on` to wait while `queue` is full.
    
    Args:
        queue: The queue to which to add the item.
        item: The item to queue.
        wait_message: The message to log while waiting.
        block_timeout: Number of seconds to wait after each `queue.put` attempt.
        kwargs: Additional arguments to `wait_on`.
    """
    def condition(item):
        """Returns True if enqueing was successful.
        """
        try:
            queue.put(item, block=True, timeout=block_timeout)
            return True
        except Full:
            return False
    wait_on(condition, item, wait_message=wait_message, **kwargs)

def enqueue_all(iterable, queue, timeout, fail_callback):
    """Enqueue all items in `iterable`, using `wait_on` to wait while `queue`
    is full.
    
    Args:
        iterable: Iterable of items to queue.
        queue: The queue to which to add the item.
        timeout: Number of seconds to wait after each `queue.put` attempt.
        fail_callback: Function called (or Exception raised) after timeout.
    
    Returns:
        The number of items queued.
    """
    num_items = 0
    def condition(item):
        """Returns True if enqueing was successful.
        """
        try:
            queue.put(item, block=True, timeout=RETRY_INTERVAL)
            return True
        except Full:
            return False
    for item in iterable:
        wait_on(
            condition, item,
            wait_message="Main process waiting to queue item {}",
            timeout=timeout,
            fail_callback=fail_callback)
        num_items += 1
    return num_items

def dequeue(
        queue, wait_message="Waiting to dequeue item {}",
        block_timeout=RETRY_INTERVAL, **kwargs):
    """Dequeue an item, using `wait_on` to wait while `queue` is empty.
    """
    def condition():
        """Returns an item from the queue, or False if the queue is empty.
        """
        try:
            return queue.get(block=True, timeout=block_timeout)
        except Empty:
            return False
    return wait_on(condition, wait_message=wait_message, **kwargs)

def kill(process, retcode, timeout):
    """Kill a process if it fails to terminate on its own.
    """
    if retcode <= 1:
        wait_on_process(process, timeout, terminate=True)
    elif process.is_alive():
        process.terminate()



# coding: utf-8
"""Collect statistics to use in the QC report.
"""
import re
from atropos.util import (
    CountingDict, NestedDict, Histogram, Mergeable, Summarizable, ordered_dict, 
    qual2int)

DEFAULT_TILE_KEY_REGEXP = r"^(?:[^\:]+\:){4}([^\:]+)"
"""Regexp for the default Illumina read name format."""

class PositionDicts(Mergeable, Summarizable):
    """A sequence of dicts, one for each position in a sequence.
    
    Args:
        is_qualities: Whether values are base qualities.
        quality_base: Base for quality values.
    """
    def __init__(self, is_qualities=False, quality_base=33):
        self.dicts = []
        self.is_qualities = is_qualities
        self.quality_base = quality_base
    
    def __getitem__(self, idx):
        if idx >= len(self.dicts):
            self.extend(idx+1)
        return self.dicts[idx]
    
    def extend(self, size):
        """Extend the number of bases to `size`.
        """
        diff = size - len(self.dicts)
        if diff > 0:
            for _ in range(diff):
                self.dicts.append(self.dict_class())
    
    def merge(self, other):
        if not isinstance(other, BaseCountingDicts):
            raise ValueError(
                "Cannot merge object of type {}".format(type(other)))
        other_len = len(other.dicts)
        min_len = min(len(self.dicts), other_len)
        for i in range(min_len):
            self.dicts[i].merge(other.dicts[i])
        if other_len > min_len:
            self.dicts.extend(other.dicts[min_len:other_len])
    
    def summarize(self):
        raise NotImplementedError()

class BaseCountingDicts(PositionDicts):
    """A PositionDicts in which the items are CountingDicts that count the
    number of items associated with each nucleotide base.
    """
    dict_class = CountingDict
    
    def summarize(self):
        """Flatten into a table with N rows (where N is the size of the
        sequence) and the columns are counts by nucleotide.
        
        Returns:
            A tuple of (columns, [rows]), where each row is
            (position, (base_counts...))
        """
        keys = set()
        for dict_item in self.dicts:
            keys.update(dict_item.keys())
        if self.is_qualities:
            keys = tuple(sorted(keys))
            columns = tuple(qual2int(k, self.quality_base) for k in keys)
        else:
            acgt = ('A','C','G','T')
            n_val = ('N',)
            columns = keys = acgt + tuple(keys - set(acgt + n_val)) + n_val
        return dict(
            columns=columns,
            rows=ordered_dict(
                (idx, tuple(dict_item.get(key, 0) for key in keys))
                for idx, dict_item in enumerate(self.dicts, 1)))

class BaseNestedDicts(PositionDicts):
    """A PositionDicts in which items are NestedDicts.
    """
    dict_class = NestedDict
    
    def summarize(self):
        """Flatten into a table of N*K rows, where N is the sequence size and
        K is the union of keys in the nested dicts, and the columns are counts
        by nucleotide.
        """
        keys1 = set()
        keys2 = set()
        for dict1 in self.dicts:
            keys1.update(dict1.keys())
            for dict2 in dict1.values():
                keys2.update(dict2.keys())
        keys1 = tuple(sorted(keys1))
        keys2 = tuple(sorted(keys2))
        if self.is_qualities:
            columns = tuple(qual2int(k, self.quality_base) for k in keys2)
        else:
            columns = keys2
        return dict(
            columns=columns,
            columns2=keys1,
            rows=ordered_dict(
                (idx, ordered_dict(
                    (key1, tuple(dict_item[key1].get(key2, 0) for key2 in keys2))
                    for key1 in keys1))
                for idx, dict_item in enumerate(self.dicts, 1)))

class ReadStatistics(object):
    """Accumulates statistics on sequencing reads.
    
    Args:
        qualities: Whether to collect base quality statistics.
        tiles: Whether to collect tile-level statistics. If True, the default
            regular expression is used, otherwise must be a regular expression
            string or compiled re for extracting the tile ID from the read name.
            Only applies to Illumina sequences.
    """
    def __init__(self, qualities=None, quality_base=33, tiles=None):
        # max read length
        self.max_read_len = 0
        # read count
        self.count = 0
        # read length distribution
        self.sequence_lengths = Histogram()
        # per-sequence GC percentage
        self.sequence_gc = Histogram()
        # per-position base composition
        self.bases = BaseCountingDicts()
        
        # whether to collect base quality stats
        self.qualities = qualities
        self.quality_base = quality_base
        self.tile_key_regexp = None
        self.sequence_qualities = None
        self.base_qualities = None
        self.tile_base_qualities = None
        
        if qualities:
            tile_key_regexp = DEFAULT_TILE_KEY_REGEXP if tiles is True else tiles
            if isinstance(tile_key_regexp, str):
                tile_key_regexp = re.compile(tile_key_regexp)
            self.tile_key_regexp = tile_key_regexp
            self._init_qualities()
        
        # cache of computed values
        self._cache = {}
    
    def _init_qualities(self):
        # per-sequence mean qualities
        self.sequence_qualities = Histogram()
        # per-position quality composition
        self.base_qualities = BaseCountingDicts(
            is_qualities=True, quality_base=self.quality_base)
        if self.tile_key_regexp:
            self.tile_base_qualities = BaseNestedDicts(
                is_qualities=True, quality_base=self.quality_base)
            self.tile_sequence_qualities = NestedDict()
    
    # These are attributes that are computed on the fly. If called by name
    # (without leading '_'), __getattr__ uses the method to compute the value
    # if it is not already cached; on subsequent calls, the cached value is
    # returned.
    
    def _gc_pct(self):
        return (
            sum(base['C'] + base['G'] for base in self.bases) /
            self.total_bases)
    
    def _total_bases(self):
        return sum(
            length * count
            for base in self.bases
            for length, count in base.items())
    
    def __getattr__(self, name):
        if name not in self._cache:
            func_name = '_' + name
            if not hasattr(self, func_name):
                raise ValueError("No function named {}".format(func_name))
            func = getattr(self, func_name)
            self._cache[name] = func()
        return self._cache[name]
    
    @property
    def track_tiles(self):
        """Whether tile statistics are being accumulated.
        """
        return self.qualities and self.tile_key_regexp is not None
    
    def collect_record(self, record):
        """Collect stats on a single sequence record.
        """
        if self.qualities is None and record.qualities:
            self.qualities = True
            self._init_qualities()
        
        seq = record.sequence
        seqlen = len(seq)
        
        self.count += 1
        self.sequence_lengths[seqlen] += 1
        
        if seqlen > 0:
            gc_pct = round((seq.count('C') + seq.count('G')) * 100 / seqlen)
            self.sequence_gc[gc_pct] += 1
            
            if seqlen > self.max_read_len:
                self._extend_bases(seqlen)
            
            quals = tile = None
            
            if self.qualities:
                quals = record.qualities
                # mean read quality
                # NOTE: we use round here, as opposed to FastQC which uses
                # floor, resulting in slightly different quality profiles
                meanqual = round(
                    sum(ord(q) - self.quality_base for q in quals) / seqlen)
                self.sequence_qualities[meanqual] += 1
                # tile ID
                if self.track_tiles:
                    tile_match = self.tile_key_regexp.match(record.name)
                    if tile_match:
                        tile = tile_match.group(1)
                        self.tile_sequence_qualities[tile][meanqual] += 1
                    else:
                        raise ValueError("{} did not match {}".format(
                            self.tile_key_regexp, record.name))
        
            # per-base nucleotide and quality composition
            for i, (base, qual) in enumerate(zip(seq, quals)):
                self.add_base(i, base, qual, tile)
        
        # TODO: positional k-mer profiles
    
    def collect(self, read1, read2=None):
        """Collect statistics on a pair of reads.
        """
        raise NotImplementedError()
    
    def add_base(self, i, base, qual=None, tile=None):
        """Add per-base information.
        
        Args:
            i: Position
            base: Nucleotide
            qual: Quality
            tile: Tile ID
        """
        self.bases[i][base] += 1
        if qual:
            self.base_qualities[i][qual] += 1
            if tile:
                self.tile_base_qualities[i][tile][qual] += 1
    
    def _extend_bases(self, new_size):
        self.bases.extend(new_size)
        if self.qualities:
            self.base_qualities.extend(new_size)
            if self.track_tiles:
                self.tile_base_qualities.extend(new_size)
    
    def summarize(self):
        """Returns a summary dict.
        """
        summary = dict(
            counts=self.count,
            lengths=self.sequence_lengths.summarize(),
            gc=self.sequence_gc.summarize(),
            bases=self.bases)
        if self.sequence_qualities:
            summary['qualities'] = self.sequence_qualities
        if self.base_qualities:
            summary['base_qualities'] = self.base_qualities
        if self.track_tiles:
            summary['tile_base_qualities'] = self.tile_base_qualities
            summary['tile_sequence_qualities'] = self.tile_sequence_qualities
        return summary

class SingleEndReadStatistics(ReadStatistics):
    """ReadStatistics for single-end data.
    """
    def collect(self, read1, read2=None):
        self.collect_record(read1)
    
    def summarize(self):
        return dict(read1=super().summarize())

class PairedEndReadStatistics(object):
    """ReadStatistics for paired-end data.
    """
    def __init__(self, **kwargs):
        self.read1 = ReadStatistics(**kwargs)
        self.read2 = ReadStatistics(**kwargs)
    
    def collect(self, read1, read2):
        """Collect statistics on a pair of reads.
        """
        self.read1.collect_record(read1)
        self.read2.collect_record(read2)
    
    def summarize(self):
        """Returns a summary dict.
        """
        return dict(
            read1=self.read1.summarize(),
            read2=self.read2.summarize())



"""Command-line interface for the error command.
"""
from atropos.commands.cli import BaseCommandParser, writeable_file
from atropos.io import STDOUT

class CommandParser(BaseCommandParser):
    name = 'error'
    usage = """
atropos error-se input.fastq
atropos error -pe in1.fq -pe2 in2.fq
"""
    description = """
Estimate the sequencing error rate. Use this command to help determine
the quality of your data, and to decide the value for the max
error rate (-e) parameter. Normal error for an Illumina experiment
is around 1% or less. We recommend setting -e to 10x the empirical
error rate."""
    
    def add_command_options(self):
        parser = self.parser
        parser.set_defaults(
            max_reads=10000,
            counter_magnitude="K",
            report_formats=['txt'])
        group = self.add_group("Error Estimation")
        group.add_argument(
            "-a",
            "--algorithm",
            choices=('quality', 'shadow'), default="quality",
            help="Method for estimating error rates; quality = base qualities, "
                 "shadow = shadow regression. Be advised that the 'shadow' "
                 "method is incredibly slow.")
        group.add_argument(
            "-m",
            "--max-bases",
            type=int, default=None,
            help="Maximum number of bases to use in the error calculation, "
                 "starting from the 5' end of the read.")
        
        group = self.add_group("Output")
        group.add_argument(
            "-o",
            "--output",
            type=writeable_file, default=STDOUT,
            help="File in which to write the summary of the estimated error "
                 "rates. (stdout)")
    
    def validate_command_options(self, options):
        options.report_file = options.output



# coding: utf-8
"""
Colorspace conversion routines.

Inspired by agapython/util/Dibase.py from Corona lite,
but reimplemented to avoid licensing issues.

Encoding Table

  A C G T
A 0 1 2 3
C 1 0 3 2
G 2 3 0 1
T 3 2 1 0
"""
def _initialize_dicts():
    """
    Create the colorspace encoding and decoding dictionaries.
    """
    enc = {}
    for i, char1 in enumerate("ACGT"):
        enc['N' + char1] = '4'
        enc[char1 + 'N'] = '4'
        enc['.' + char1] = '4'
        enc[char1 + '.'] = '4'
        for j, char2 in enumerate("ACGT"):
            # XOR of nucleotides gives color
            enc[char1 + char2] = chr(ord('0') + (i ^ j))
    enc.update({ 'NN': '4', 'N.': '4', '.N': '4', '..': '4'})

    dec = {}
    for i, char1 in enumerate("ACGT"):
        dec['.' + str(i)] = 'N'
        dec['N' + str(i)] = 'N'
        dec[char1 + '4'] = 'N'
        dec[char1 + '.'] = 'N'
        for j, char2 in enumerate("ACGT"):
            # XOR of nucleotides gives color
            dec[char1 + chr(ord('0') + (i ^ j))] = char2
    dec['N4'] = 'N'

    return (enc, dec)

ENCODE, DECODE = _initialize_dicts()

def encode(nucs):
    """Given a sequence of nucleotides, convert them to colorspace. Only
    uppercase characters are allowed.
    
    Examples:
        >>> encode("ACGGTC")
        "A13012"
    """
    if not nucs:
        return nucs
    encoded = nucs[0:1]
    for idx in range(len(nucs) - 1):
        encoded += ENCODE[nucs[idx:idx+2]]
    return encoded

def decode(colors):
    """Decode a sequence of colors to nucleotide space. The first character in
    `colors` must be a nucleotide. Only uppercase characters are allowed.
    
    Examples:
        >>> decode("A13012")
        "ACGGTC"
    """
    if len(colors) < 2:
        return colors
    result = base = colors[0]
    for col in colors[1:]:
        base = DECODE[base + col]
        result += base
    return result



# coding: utf-8
"""Top-level Atropos package.
"""
import os
import sys

from atropos._version import get_versions
__version__ = get_versions()['version']
del get_versions

class AtroposError(Exception):
    """Base class for Atropos-specific errors.
    """
    pass

def check_importability():  # pragma: no cover
    """Check that cython modules haev been compile.
    """
    try:
        import atropos.align._align # pylint: disable=unused-variable
    except ImportError as err:
        if 'undefined symbol' in str(err):
            print("""
ERROR: A required extension module could not be imported because it is
incompatible with your system. A quick fix is to recompile the extension
modules with the following command:

    {0} setup.py build_ext -i

See the documentation for alternative ways of installing the program.

The original error message follows.
""".format(sys.executable))
        raise



# coding: utf-8
from atropos.adapters import Adapter, ColorspaceAdapter, PREFIX, BACK
from atropos.commands.trim.modifiers import AdapterCutter
from atropos.io.seqio import ColorspaceSequence, Sequence

def test_cs_5p():
    read = ColorspaceSequence("name", "0123", "DEFG", "T")
    adapter = ColorspaceAdapter("CG", PREFIX, 0.1)
    cutter = AdapterCutter([adapter])
    trimmed_read = cutter(read)
    # no assertion here, just make sure the above code runs without
    # an exception


def test_statistics():
    read = Sequence('name', 'AAAACCCCAAAA')
    adapters = [Adapter('CCCC', BACK, 0.1)]
    cutter = AdapterCutter(adapters, times=3)
    trimmed_read = cutter(read)
    # TODO make this a lot simpler
    trimmed_bp = 0
    for adapter in adapters:
        for d in (adapter.lengths_front, adapter.lengths_back):
            trimmed_bp += sum(seqlen * count for (seqlen, count) in d.items())
    assert trimmed_bp <= len(read), trimmed_bp



# coding: utf-8
# TODO
# test with the --output option
# test reading from standard input
from io import StringIO
import os
from pytest import raises
import sys
from atropos.commands import execute_cli, get_command
from .utils import (
    run, files_equal, datapath, cutpath, redirect_stderr, temporary_path)

def test_example():
    run('-N -b ADAPTER', 'example.fa', 'example.fa')

def test_small():
    run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'small.fastq')

def test_empty():
    '''empty input'''
    run('-a TTAGACATATCTCCGTCG', 'empty.fastq', 'empty.fastq')

def test_newlines():
    '''DOS/Windows newlines'''
    run('-e 0.12 -b TTAGACATATCTCCGTCG', 'dos.fastq', 'dos.fastq')

def test_lowercase():
    '''lowercase adapter'''
    run('-b ttagacatatctccgtcg', 'lowercase.fastq', 'small.fastq')


def test_rest():
    '''-r/--rest-file'''
    with temporary_path('rest.tmp') as rest_tmp:
        run(['-b', 'ADAPTER', '-N', '-r', rest_tmp], "rest.fa", "rest.fa")
        assert files_equal(datapath('rest.txt'), rest_tmp)


def test_restfront():
    with temporary_path("rest.txt") as path:
        run(['-g', 'ADAPTER', '-N', '-r', path], "restfront.fa", "rest.fa")
        assert files_equal(datapath('restfront.txt'), path)


def test_discard():
    '''--discard'''
    run("-b TTAGACATATCTCCGTCG --discard", "discard.fastq", "small.fastq")


def test_discard_untrimmed():
    '''--discard-untrimmed'''
    run('-b CAAGAT --discard-untrimmed', 'discard-untrimmed.fastq', 'small.fastq')


def test_plus():
    '''test if sequence name after the "+" is retained'''
    run("-e 0.12 -b TTAGACATATCTCCGTCG", "plus.fastq", "plus.fastq")


def test_extensiontxtgz():
    '''automatic recognition of "_sequence.txt.gz" extension'''
    run("-b TTAGACATATCTCCGTCG", "s_1_sequence.txt", "s_1_sequence.txt.gz")


def test_format():
    '''the -f/--format parameter'''
    run("-f fastq -b TTAGACATATCTCCGTCG", "small.fastq", "small.myownextension")


def test_minimum_length():
    '''-m/--minimum-length'''
    run("-c -m 5 -a 330201030313112312", "minlen.fa", "lengths.fa")


def test_too_short():
    '''--too-short-output'''
    run("-c -m 5 -a 330201030313112312 --too-short-output tooshort.tmp.fa", "minlen.fa", "lengths.fa")
    assert files_equal(datapath('tooshort.fa'), "tooshort.tmp.fa")
    os.remove('tooshort.tmp.fa')


def test_too_short_no_primer():
    '''--too-short-output and --trim-primer'''
    run("-c -m 5 -a 330201030313112312 --trim-primer --too-short-output tooshort.tmp.fa", "minlen.noprimer.fa", "lengths.fa")
    assert files_equal(datapath('tooshort.noprimer.fa'), "tooshort.tmp.fa")
    os.remove('tooshort.tmp.fa')


def test_maximum_length():
    '''-M/--maximum-length'''
    run("-c -M 5 -a 330201030313112312", "maxlen.fa", "lengths.fa")


def test_too_long():
    '''--too-long-output'''
    run("-c -M 5 --too-long-output toolong.tmp.fa -a 330201030313112312", "maxlen.fa", "lengths.fa")
    assert files_equal(datapath('toolong.fa'), "toolong.tmp.fa")
    os.remove('toolong.tmp.fa')


def test_length_tag():
    '''454 data; -n and --length-tag'''
    run("-n 3 -e 0.1 --length-tag length= " \
        "-b TGAGACACGCAACAGGGGAAAGGCAAGGCACACAGGGGATAGG "\
        "-b TCCATCTCATCCCTGCGTGTCCCATCTGTTCCCTCCCTGTCTCA", '454.fa', '454.fa')

def test_overlap_a():
    '''-O/--overlap with -a (-c omitted on purpose)'''
    run("-O 10 -a 330201030313112312 -e 0.0 -N", "overlapa.fa", "overlapa.fa")

def test_overlap_b():
    '''-O/--overlap with -b'''
    run("-O 10 -b TTAGACATATCTCCGTCG -N", "overlapb.fa", "overlapb.fa")

def test_qualtrim():
    '''-q with low qualities'''
    run("-q 10 -a XXXXXX", "lowqual.fastq", "lowqual.fastq")

def test_qualbase():
    '''-q with low qualities, using ascii(quality+64) encoding'''
    run("-q 10 --quality-base 64 -a XXXXXX", "illumina64.fastq", "illumina64.fastq")

def test_quality_trim_only():
    '''only trim qualities, do not remove adapters'''
    run("-q 10 --quality-base 64", "illumina64.fastq", "illumina64.fastq")

def test_twoadapters():
    '''two adapters'''
    run("-a AATTTCAGGAATT -a GTTCTCTAGTTCT", "twoadapters.fasta", "twoadapters.fasta")

def test_polya():
    '''poly-A tails'''
    run("-m 24 -O 10 -a AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA", "polya.fasta", "polya.fasta")

def test_polya_brace_notation():
    '''poly-A tails'''
    run("-m 24 -O 10 -a A{35}", "polya.fasta", "polya.fasta")

def test_mask_adapter():
    '''mask adapter with N (reads maintain the same length)'''
    run("-b CAAG -n 3 --mask-adapter", "anywhere_repeat.fastq", "anywhere_repeat.fastq")

def test_gz_multiblock():
    '''compressed gz file with multiple blocks (created by concatenating two .gz files)'''
    run("-b TTAGACATATCTCCGTCG", "small.fastq", "multiblock.fastq.gz")

def test_suffix():
    '''-y/--suffix parameter, combined with _F3'''
    run("-c -e 0.12 -a 1=330201030313112312 -y _my_suffix_{name} --strip-f3", "suffix.fastq", "solid.csfasta", qualfile='solid.qual')

def test_read_wildcard():
    '''test wildcards in reads'''
    run("--match-read-wildcards -b ACGTACGT", "wildcard.fa", "wildcard.fa")

def test_adapter_wildcard():
    '''wildcards in adapter'''
    for adapter_type, expected in (
            ("-a", "wildcard_adapter.fa"),
            ("-b", "wildcard_adapter_anywhere.fa")):
        with temporary_path("wildcardtmp.txt") as wildcardtmp:
            run("--wildcard-file {0} {1} ACGTNNNACGT".format(wildcardtmp, adapter_type),
                expected, "wildcard_adapter.fa")
            with open(wildcardtmp) as wct:
                lines = wct.readlines()
            lines = [ line.strip() for line in lines ]
            assert lines == ['AAA 1', 'GGG 2', 'CCC 3b', 'TTT 4b']

def test_wildcard_N():
    '''test 'N' wildcard matching with no allowed errors'''
    run("-e 0 -a GGGGGGG --match-read-wildcards", "wildcardN.fa", "wildcardN.fa")

def test_illumina_adapter_wildcard():
    run("-a VCCGAMCYUCKHRKDCUBBCNUWNSGHCGU", "illumina.fastq", "illumina.fastq.gz")

def test_adapter_front():
    '''test adapter in front'''
    run("--front ADAPTER -N", "examplefront.fa", "example.fa")

def test_literal_N():
    '''test matching literal 'N's'''
    run("-N -e 0.2 -a NNNNNNNNNNNNNN", "trimN3.fasta", "trimN3.fasta")

def test_literal_N2():
    run("-N -O 1 -g NNNNNNNNNNNNNN", "trimN5.fasta", "trimN5.fasta")

def test_literal_N_brace_notation():
    '''test matching literal 'N's'''
    run("-N -e 0.2 -a N{14}", "trimN3.fasta", "trimN3.fasta")

def test_literal_N2_brace_notation():
    run("-N -O 1 -g N{14}", "trimN5.fasta", "trimN5.fasta")

def test_anchored_front():
    run("-g ^FRONTADAPT -N", "anchored.fasta", "anchored.fasta")

def test_anchored_front_ellipsis_notation():
    run("-a FRONTADAPT... -N", "anchored.fasta", "anchored.fasta")

def test_anchored_back():
    run("-a BACKADAPTER$ -N", "anchored-back.fasta", "anchored-back.fasta")

def test_anchored_back_no_indels():
    run("-a BACKADAPTER$ -N --no-indels", "anchored-back.fasta", "anchored-back.fasta")


def test_no_indels():
    run('-a TTAGACATAT -g GAGATTGCCA --no-indels', 'no_indels.fasta', 'no_indels.fasta')


def test_issue_46():
    '''issue 46 - IndexError with --wildcard-file'''
    with temporary_path("wildcardtmp.txt") as wildcardtmp:
        run("--anywhere=AACGTN --wildcard-file={0}".format(wildcardtmp), "issue46.fasta", "issue46.fasta")

def test_strip_suffix():
    run("--strip-suffix _sequence -a XXXXXXX", "stripped.fasta", "simple.fasta")


def test_info_file():
    # The true adapter sequence in the illumina.fastq.gz data set is
    # GCCTAACTTCTTAGACTGCCTTAAGGACGT (fourth base is different)
    #
    with temporary_path("infotmp.txt") as infotmp:
        run(["--info-file", infotmp, '-a', 'adapt=GCCGAACTTCTTAGACTGCCTTAAGGACGT'], "illumina.fastq", "illumina.fastq.gz")
        assert files_equal(cutpath('illumina.info.txt'), infotmp)


def test_info_file_times():
    with temporary_path("infotmp.txt") as infotmp:
        run(["--info-file", infotmp, '--times', '2', '-a', 'adapt=GCCGAACTTCTTA', '-a', 'adapt2=GACTGCCTTAAGGACGT'], "illumina5.fastq", "illumina5.fastq")
        assert files_equal(cutpath('illumina5.info.txt'), infotmp)


def test_info_file_fasta():
    with temporary_path("infotmp.txt") as infotmp:
        # Just make sure that it runs
        run(['--info-file', infotmp, '-a', 'TTAGACATAT', '-g', 'GAGATTGCCA', '--no-indels'], 'no_indels.fasta', 'no_indels.fasta')


def test_named_adapter():
    run("-a MY_ADAPTER=GCCGAACTTCTTAGACTGCCTTAAGGACGT", "illumina.fastq", "illumina.fastq.gz")


def test_adapter_with_U():
    run("-a GCCGAACUUCUUAGACUGCCUUAAGGACGU", "illumina.fastq", "illumina.fastq.gz")


def test_no_trim():
    ''' --no-trim '''
    run("--no-trim --discard-untrimmed -a CCCTAGTTAAAC", 'no-trim.fastq', 'small.fastq')


def test_bzip2():
    '''test bzip2 support'''
    run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'small.fastq.bz2')


try:
    import lzma

    def test_xz():
        '''test xz support'''
        run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'small.fastq.xz')
except ImportError:
    pass


def test_qualfile_only():
    with raises(SystemExit), redirect_stderr():
        execute_cli(['-sq', datapath('E3M.qual')])


def test_no_args():
    with redirect_stderr():
        assert execute_cli() != 0


def test_two_fastqs():
    with raises(SystemExit), redirect_stderr():
        execute_cli(['-pe1', datapath('paired.1.fastq'), '-pe2', datapath('paired.2.fastq')])


def test_anchored_no_indels():
    '''anchored 5' adapter, mismatches only (no indels)'''
    run('-g ^TTAGACATAT --no-indels -e 0.1', 'anchored_no_indels.fasta', 'anchored_no_indels.fasta')


def test_anchored_no_indels_wildcard_read():
    '''anchored 5' adapter, mismatches only (no indels), but wildcards in the read count as matches'''
    run('-g ^TTAGACATAT --match-read-wildcards --no-indels -e 0.1', 'anchored_no_indels_wildcard.fasta', 'anchored_no_indels.fasta')


def test_anchored_no_indels_wildcard_adapt():
    '''anchored 5' adapter, mismatches only (no indels), but wildcards in the adapter count as matches'''
    run('-g ^TTAGACANAT --no-indels -e 0.1', 'anchored_no_indels.fasta', 'anchored_no_indels.fasta')


def test_unconditional_cut_front():
    run('-u 5', 'unconditional-front.fastq', 'small.fastq')


def test_unconditional_cut_back():
    run('-u -5', 'unconditional-back.fastq', 'small.fastq')


def test_unconditional_cut_both():
    run('-u -5 -u 5', 'unconditional-both.fastq', 'small.fastq')


def test_untrimmed_output():
    with temporary_path('untrimmed.tmp.fastq') as tmp:
        run(['-a', 'TTAGACATATCTCCGTCG', '--untrimmed-output', tmp], 'small.trimmed.fastq', 'small.fastq')
        assert files_equal(cutpath('small.untrimmed.fastq'), tmp)


def test_adapter_file():
    run('-a file:' + datapath('adapter.fasta'), 'illumina.fastq', 'illumina.fastq.gz')

def test_adapter_file_5p_anchored():
    run('-N -g file:' + datapath('prefix-adapter.fasta'), 'anchored.fasta', 'anchored.fasta')

def test_adapter_file_3p_anchored():
    run('-N -a file:' + datapath('suffix-adapter.fasta'), 'anchored-back.fasta', 'anchored-back.fasta')


def test_adapter_file_5p_anchored_no_indels():
    run('-N --no-indels -g file:' + datapath('prefix-adapter.fasta'), 'anchored.fasta', 'anchored.fasta')


def test_adapter_file_3p_anchored_no_indels():
    run('-N --no-indels -a file:' + datapath('suffix-adapter.fasta'), 'anchored-back.fasta', 'anchored-back.fasta')


def test_demultiplex():
    multiout = os.path.join(os.path.dirname(__file__), 'data', 'tmp-demulti.{name}.fasta')
    params = ['-a', 'first=AATTTCAGGAATT', '-a', 'second=GTTCTCTAGTTCT', '-o', multiout, '-se', datapath('twoadapters.fasta')]
    command = get_command('trim')
    result = command.execute(params)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] == 0
    assert files_equal(cutpath('twoadapters.first.fasta'), multiout.format(name='first'))
    assert files_equal(cutpath('twoadapters.second.fasta'), multiout.format(name='second'))
    assert files_equal(cutpath('twoadapters.unknown.fasta'), multiout.format(name='unknown'))
    os.remove(multiout.format(name='first'))
    os.remove(multiout.format(name='second'))
    os.remove(multiout.format(name='unknown'))


def test_max_n():
    run('--max-n 0', 'maxn0.fasta', 'maxn.fasta')
    run('--max-n 1', 'maxn1.fasta', 'maxn.fasta')
    run('--max-n 2', 'maxn2.fasta', 'maxn.fasta')
    run('--max-n 0.2', 'maxn0.2.fasta', 'maxn.fasta')
    run('--max-n 0.4', 'maxn0.4.fasta', 'maxn.fasta')


def test_quiet_is_quiet():
    captured_standard_output = StringIO()
    captured_standard_error = StringIO()
    try:
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = captured_standard_output
        sys.stderr = captured_standard_error
        execute_cli(['-o', '/dev/null', '--quiet', '-a', 'XXXX', '-se', datapath('illumina.fastq.gz')])
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
    print(captured_standard_output.getvalue())
    assert captured_standard_output.getvalue() == ''
    assert captured_standard_error.getvalue() == ''


def test_nextseq():
    run('--nextseq-trim 22', 'nextseq.fastq', 'nextseq.fastq')


def test_linked():
    run('-a AAAAAAAAAA...TTTTTTTTTT', 'linked.fasta', 'linked.fasta')


def test_fasta():
    run('-a TTAGACATATCTCCGTCG', 'small.fasta', 'small.fastq')

def test_custom_bisulfite_1():
    run('-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,1,1', 'small.fastq', 'small.fastq')
def test_custom_bisulfite_2():
    run('-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 15,15,1,1', 'small_mincut1.fastq', 'small.fastq')
def test_custom_bisulfite_3():
    run('-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,1,0', 'small_mincut2.fastq', 'small.fastq')
def test_custom_bisulfite_4():
    run('-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,0,0', 'small_mincut3.fastq', 'small.fastq')



"""Detect adapter sequences directly from reads based on kmer frequency.
"""
from collections import defaultdict
import logging
import math
import re
from atropos.align import Aligner, SEMIGLOBAL
from atropos.commands.base import BaseCommandRunner
from atropos.util import (
    reverse_complement, sequence_complexity, enumerate_range)

# TODO: Test whether using rc=True in parse_known_contaminants is as fast
# and as accurate as testing both the forward and reverse complement
# read sequence against only the forward-orientation known contaminants.
# Also, offer an option of whether to test the reverse complement, with
# the default being false.

# TODO: Look at CountTable in khmer 2.1. Otherwise move to pymer.
# pymer https://github.com/kdmurray91/pymer/tree/master/pymer

# TODO: Import kmers from FastQC output.

# TODO: In KnownContaminantDetector, accept template sequences with wildcards
# to match against.

# TODO: Re-download sequencing_adapters.fa if it has been updated since last
# download.

class CommandRunner(BaseCommandRunner):
    name = 'detect'
    
    def __call__(self):
        kmer_size = self.kmer_size or 12
        n_reads = self.max_reads
        overrep_cutoff = 100
        include = self.include_contaminants or "all"
        known_contaminants = None
        if include != 'unknown':
            known_contaminants = self.load_known_adapters()
        
        detector = self.detector
        if not detector:
            if known_contaminants and include == 'known':
                detector = 'known'
            elif n_reads <= 50000:
                detector = 'heuristic'
            else:
                detector = 'khmer'
        
        if detector == 'known':
            logging.getLogger().debug(
                "Detecting contaminants using the known-only algorithm")
            detector_class = KnownContaminantDetector
        elif detector == 'heuristic':
            logging.getLogger().debug(
                "Detecting contaminants using the heuristic algorithm")
            detector_class = HeuristicDetector
        elif detector == 'khmer':
            logging.getLogger().debug(
                "Detecting contaminants using the kmer-based algorithm")
            detector_class = KhmerDetector
        
        input_names = self.input_names
        detector_args = dict(
            kmer_size=kmer_size, n_reads=n_reads, overrep_cutoff=overrep_cutoff,
            known_contaminants=known_contaminants)
        if self.paired:
            detector = PairedDetector(detector_class, **detector_args)
        else:
            detector = detector_class(**detector_args)
            input_names = input_names[0]
        
        self.summary['detect'] = dict(
            kmer_size=kmer_size,
            n_reads=n_reads,
            include=include,
            detector=detector)
        
        logging.getLogger().info(
            "Detecting adapters and other potential contaminant "
            "sequences based on %d-mers in %d reads", kmer_size, n_reads)
        
        detector.consume_all_batches(self)
        
        return 0

class Match(object):
    """A contaminant match.
    
    Args:
        seq_or_contam: The matched sequence.
        count: The number of matches.
        names: The name(s) of matching contaminant(s).
        match_frac: The fraction of contaminant kmers that match.
        match_frac2: The fraction of sequence kmers that match.
        reads: The number of reads with the contaminant.
    """
    def __init__(
            self, seq_or_contam, count=0, names=None, match_frac=None,
            match_frac2=None, reads=None):
        if isinstance(seq_or_contam, ContaminantMatcher):
            self.seq = seq_or_contam.seq
            self.count = seq_or_contam.matches
            self.names = tuple(seq_or_contam.names)
            self.known_seqs = [seq_or_contam.seq]
        else:
            self.seq = seq_or_contam
            self.count = count
            self.names = names
            self.known_seqs = None
        self.match_frac = match_frac
        self.match_frac2 = match_frac2
        self.abundance = None
        self.longest_match = None
        if reads:
            self.set_longest_match(reads)
    
    def __len__(self):
        return len(self.seq)
    
    def __repr__(self):
        if self.is_known:
            return "{} => {} ({}))".format(
                self.seq, self.names, self.known_seqs)
        else:
            return self.seq
    
    @property
    def seq_complexity(self):
        """The complexity of the sequence (0=homopolymer, 2=random).
        """
        return sequence_complexity(self.seq)
    
    @property
    def count_is_frequency(self):
        """Whether `self.count` represents a frequency.
        """
        return isinstance(self.count, float)
    
    def set_contaminant(self, contam, match_frac, match_frac2=None):
        """Set the known contaminant from a :class:`ContaminantMatcher`.
        """
        self.set_known(contam.names, [contam.seq], match_frac, match_frac2)
    
    def set_known(self, names, seqs, match_frac, match_frac2=None):
        """Set the known contaminant.
        """
        self.names = names
        self.known_seqs = seqs
        self.match_frac = match_frac
        self.match_frac2 = match_frac2
    
    @property
    def is_known(self):
        """Whether the matching sequence is a known contaminant.
        """
        return self.known_seqs is not None
    
    def set_longest_match(self, sequences):
        """Set the longest matching sequence from a set of sequences.
        """
        for seq in sequences:
            idx = seq.index(self.seq)
            seqlen = len(self.seq) - idx
            if self.longest_match is None or self.longest_match[1] < seqlen:
                self.longest_match = (seq[idx:], seqlen)
    
    def estimate_abundance(self, read_sequences):
        """Determine whether this match's sequence is within 'seq' by simple
        exact string comparison.
        """
        self.abundance = sum(
            1 for read_seq in read_sequences
            if self.seq in read_seq)

class ContaminantMatcher(object):
    """Matches a known contaminant against other sequences.
    
    Args:
        seq: The contaminant sequence.
        names: Sequence of names for the contaminant.
        kmer_size: kmer size.
    """
    def __init__(self, seq, names, kmer_size):
        self.seq = seq
        self.names = names
        self.kmers = set(
            seq[i:(i+kmer_size)]
            for i in range(len(seq) - kmer_size + 1))
        self.n_kmers = len(self.kmers)
        self.kmer_size = kmer_size
        self.matches = 0
        
    def match(self, seq, seqrc):
        """Returns (num_matches, num_contam_kmers, num_seq_kmers).
        
        Args:
            seq: The sequence to match.
            seqrc: The reverse complement of `seq`.
        
        Returns:
            Tuple (f1, f2, seq), where f1 is the fraction of contaminant kmers
            that match, f2 is the fraction of sequence kmers that match, and
            seq is the best matching sequence (either `seq` or `seqrc`).
        """
        fw_kmers = set(
            seq[i:(i+self.kmer_size)]
            for i in range(len(seq) - self.kmer_size + 1))
        fw_matches = float(len(self.kmers & fw_kmers))
        
        rv_kmers = set(
            seqrc[i:(i+self.kmer_size)]
            for i in range(len(seqrc) - self.kmer_size + 1))
        rv_matches = float(len(self.kmers & rv_kmers))
        
        if fw_matches >= rv_matches:
            n_matches = fw_matches
            kmers = fw_kmers
            compare_seq = seq
        else:
            n_matches = rv_matches
            kmers = rv_kmers
            compare_seq = seqrc
        
        self.matches += n_matches
        match_frac1 = match_frac2 = 0
        if self.n_kmers > 0:
            match_frac1 = n_matches / self.n_kmers
        if len(kmers) > 0:
            match_frac2 = n_matches / len(kmers)
        return (match_frac1, match_frac2, compare_seq)

def create_contaminant_matchers(contaminants, kmer_size):
    """Create :class:`ContaminantMatcher`s from sequences.
    
    Args:
        contaminants: A dict of {seq:names}.
        kmer_size: The kmer size.
    
    Returns:
        A list of :class:`ContaminantMatcher`s.
    """
    return [
        ContaminantMatcher(seq, names, kmer_size)
        for seq, names in contaminants.iter_sequences()
    ]

POLY_A = re.compile('A{8,}.*|A{2,}$')

class Detector(object):
    """Base class for contaminant detectors.
    
    Args:
        kmer_size: Size of kmers to match.
        n_reads: Number of reads to sample.
        overrep_cutoff: Degree of overrepresentation required for a kmer to be
            considered as a contaminant.
        known_contaminant: :class:`ContaminantMatcher`s to match against.
    """
    def __init__(
            self, kmer_size=12, n_reads=10000, overrep_cutoff=100,
            known_contaminants=None):
        self.kmer_size = kmer_size
        self.n_reads = n_reads
        self.overrep_cutoff = overrep_cutoff
        self.known_contaminants = known_contaminants
        self._read_length = None
        self._read_sequences = set()
        self._matches = None
    
    @property
    def min_report_freq(self):
        """The minimum contaminant frequency required for a contaminant to be
        reported.
        """
        raise NotImplementedError()
    
    def consume_all(self, reader):
        """Consume up to self.n_reads reads from the reader.
        
        Args:
            reader: Iterator over sequence records.
        """
        read = next(reader)
        self.consume_first(read)
        for read in enumerate_range(reader, 1, self.n_reads):
            self.consume(read)
    
    def consume_all_batches(self, batch_iterator):
        """Consume all reads from the specified batch_iterator. It is expected
        that the iterator was constructed with max_reads == n_reads.
        
        Args:
            batch_iterator: Iterator over batches of sequence records.
        """
        for batch_num, (batch_meta, batch) in enumerate(batch_iterator):
            if batch_meta['size'] == 0:
                continue
            if batch_num == 0:
                self.consume_first(batch[0])
                batch = batch[1:]
            for read in batch:
                self.consume(read)
        
    def consume_first(self, read):
        """Consumes the first sequence record, which determines the read length.
        """
        assert self._read_length is None
        self._read_length = len(read.sequence)
        self.consume(read)
    
    def consume(self, read):
        """Consumes a single read.
        """
        seq = self._filter_seq(read.sequence)
        if seq:
            self._read_sequences.add(seq)
    
    def _filter_seq(self, seq):
        if sequence_complexity(seq) <= 1.0:
            return None
        match = POLY_A.search(seq)
        if match:
            seq = seq[:match.start()]
        if len(seq) < self.kmer_size:
            return None
        return seq
    
    def matches(self, **kwargs):
        """Returns the current set of matches.
        """
        if self._matches is None or len(kwargs) > 0:
            self._filter_and_sort(**kwargs)
        return self._matches
    
    def _filter_and_sort(
            self, include="all", min_len=None, min_complexity=1.1,
            min_match_frac=0.1, limit=20):
        """Identify, filter, and sort contaminants.
        
        Args:
            include: Contaminants to include; 'all', 'known', or 'unknown'.
            min_len: Minimum contaminant length.
            min_complexity: Minimum sequence complexity.
            min_match_frac: Minimum fraction of matching kmers.
            limit: Maximum number of contaminants to return.
        """
        if min_len is None:
            min_len = self.kmer_size
        
        matches = self._get_contaminants()
        
        for match in matches:
            match.estimate_abundance(self._read_sequences)
        
        def _filter(match):
            if match.count < self.min_report_freq:
                return False
            if min_len and len(match) < min_len:
                return False
            if min_complexity and match.seq_complexity < min_complexity:
                return False
            if include == 'known' and not match.is_known:
                return False
            elif include == 'unknown' and match.is_known:
                return False
            if (
                    min_match_frac and match.is_known and
                    match.match_frac < min_match_frac):
                return False
            return True
        
        matches = list(filter(_filter, matches))
        matches.sort(key=lambda x: len(x) * math.log(x.count), reverse=True)
        
        if limit is not None:
            matches = matches[:limit]
        
        self._matches = matches
    
    def _get_contaminants(self):
        """Implemention of contaminant matching.
        
        Returns:
            A list of :class:`Match`es.
        """
        raise NotImplementedError()
    
    def summarize(self, outstream, name=None, **kwargs):
        """Print a summary.
        """
        header = "File: {}".format(name) if name else None
        summarize_contaminants(
            outstream, self.matches(**kwargs), self.n_reads, header)

class PairedDetector(object):
    """Detector for paired-end reads.
    """
    def __init__(self, detector_class, **kwargs):
        self.read1_detector = detector_class(**kwargs)
        self.read2_detector = detector_class(**kwargs)
    
    def consume_all(self, reader):
        """Consume all records in `reader`.
        """
        read1, read2 = next(reader)
        self.read1_detector.consume_first(read1)
        self.read2_detector.consume_first(read2)
        for read1, read2 in reader:
            self.read1_detector.consume(read1)
            self.read2_detector.consume(read2)
    
    def consume_all_batches(self, batch_iterator):
        """Consume all batches in `batch_iterator`.
        """
        for batch_num, (batch_meta, batch) in enumerate(batch_iterator):
            if batch_meta['size'] == 0:
                continue
            if batch_num == 0:
                read1, read2 = batch[0]
                self.read1_detector.consume_first(read1)
                self.read2_detector.consume_first(read2)
                batch = batch[1:]
            for read1, read2 in batch:
                self.read1_detector.consume(read1)
                self.read2_detector.consume(read2)
    
    def matches(self, **kwargs):
        """Returns the tuple (read1_matches, read2_matches).
        """
        return (
            self.read1_detector.matches(**kwargs),
            self.read2_detector.matches(**kwargs))
    
    def summarize(self, outstream, names=(None, None), **kwargs):
        """Print a summary.
        """
        name1, name2 = names
        self.read1_detector.summarize(outstream, name1, **kwargs)
        self.read2_detector.summarize(outstream, name2, **kwargs)

class KnownContaminantDetector(Detector):
    """Test known contaminants against reads. This has linear complexity and is
    more specific than the khmer matcher, but less specific than the heuristic
    matcher. It's also less sensitive since it does not try to detect unknown
    contaminants.
    
    Args:
        known_contaminants: List of :class:`ContaminantMatcher`s.
        min_match_frac: Minimum fraction of matching kmers required.
        kwargs: Additional arguments to pass to the :class:`Detector`
            constructor.
    """
    def __init__(self, known_contaminants, min_match_frac=0.5, **kwargs):
        super().__init__(known_contaminants=known_contaminants, **kwargs)
        self.min_match_frac = min_match_frac
        self._min_k = min(len(s) for s in known_contaminants.sequences)
    
    @property
    def min_report_freq(self):
        return 0.1
    
    def _filter_seq(self, seq):
        seq = super()._filter_seq(seq)
        if seq and len(seq) >= self._min_k:
            return seq
        return None
    
    def _get_contaminants(self):
        contaminant_matchers = create_contaminant_matchers(
            self.known_contaminants, self.kmer_size)
        counts = defaultdict(lambda: 0)

        for seq in self._read_sequences:
            seqrc = reverse_complement(seq)
            for contam in contaminant_matchers:
                match = contam.match(seq, seqrc)
                if match[0] > self.min_match_frac:
                    counts[contam] += 1
        
        min_count = math.ceil(
            self.n_reads * (self._read_length - self._min_k + 1) *
            self.overrep_cutoff / float(4**self._min_k))
        
        return [
            Match(c[0], match_frac=float(c[1]) / self.n_reads)
            for c in filter(
                lambda x: x[1] >= min_count,
                counts.items()
            )
        ]

class HeuristicDetector(Detector):
    """Use a heuristic iterative algorithm to arrive at likely contaminants.
    This is the most accurate algorithm overall, but it has quadratic complexity
    and becomes too slow/memory-intenstive when n_reads > 50k.
    """
    def __init__(
            self, min_freq=0.001, min_contaminant_match_frac=0.9, **kwargs):
        super(HeuristicDetector, self).__init__(**kwargs)
        self.min_freq = min_freq
        self.min_contaminant_match_frac = min_contaminant_match_frac
    
    @property
    def min_report_freq(self):
        return 0.1 * self.n_reads
    
    def _get_contaminants(self):
        def _min_count(kmer_size):
            return math.ceil(self.n_reads * max(
                self.min_freq,
                (self._read_length - kmer_size + 1) * self.overrep_cutoff /
                float(4**kmer_size)))
        
        kmer_size = self.kmer_size
        kmers = defaultdict(lambda: [0, set()])
        
        for seq in self._read_sequences:
            for i in range(len(seq) - kmer_size + 1):
                kmer = seq[i:(i+kmer_size)]
                kmers[kmer][0] += 1
                kmers[kmer][1].add(seq)
        
        prev = None
        cur = {}
        results = {}
        result_seqs = defaultdict(set)
        min_count = _min_count(kmer_size)
        
        # Identify candidate kmers for increasing values of k
        while True:
            all_seqs = set()
            for kmer, (count, seqs) in kmers.items():
                if count > min_count:
                    cur[kmer] = (count, seqs)
                    all_seqs.update(seqs)
            
            if len(all_seqs) == 0:
                break
            
            if prev:
                for kmer, (count, seqs) in prev.items():
                    if (
                            not any(seq in cur for seq in seqs) and
                            sequence_complexity(kmer) > 1.0):
                        results[kmer] = count
                        result_seqs[kmer].update(seqs)
            
            kmer_size += 1
            kmers = defaultdict(lambda: [0, set()])
            for seq in all_seqs:
                for i in range(len(seq) - kmer_size + 1):
                    kmer = seq[i:(i+kmer_size)]
                    kmers[kmer][0] += 1
                    kmers[kmer][1].add(seq)
            
            min_count = _min_count(kmer_size)
            prev = cur
            cur = {}
        
        results = list(results.items())
        
        # Now merge overlapping sequences by length and frequency to eliminate
        # redundancy in the set of candidate kmers.
        results.sort(key=lambda i: len(i[0]) * math.log(i[1]), reverse=True)
        cur = results[0]
        merged = []
        unmerged = []
        while len(results) > 1:
            seq1, count1 = results[0]
            for j in range(1, len(results)):
                seq2, count2 = results[j]
                if len(seq1) >= len(seq2) and seq2 in seq1:
                    count1 += count2
                elif seq1 in seq2:
                    # if they are close in count, keep the longer sequence
                    if count1 < (2 * count2):
                        seq1 = seq2
                    count1 += count2
                else:
                    unmerged.append(results[j])
            merged.append([seq1, count1])
            results = unmerged
            unmerged = []
        results = merged + results
        
        if len(results) == 0:
            return []
        
        # TODO: For each retained match, pull out the longest sequence that
        # matches to have a better shot of identifying long adapters that
        # appear in full very infrequently
        
        # Re-sort by frequency
        results.sort(key=lambda i: i[1], reverse=True)
        # Keep anything that's within 50% of the top hit
        # TODO: make this user-configurable?
        min_count = int(results[0][1] * 0.5)
        results = (x for x in results if x[1] >= min_count)
        # Convert to matches
        matches = [Match(x[0], x[1], reads=result_seqs[x[0]]) for x in results]
        
        if self.known_contaminants:
            # Match to known sequences
            contaminants = create_contaminant_matchers(
                self.known_contaminants, self.kmer_size)
            known = {}
            unknown = []
            
            def find_best_match(seq, best_matches, best_match_frac):
                """Find best contaminant matches to `seq`.
                """
                seqrc = reverse_complement(seq)
                for contam in contaminants:
                    match_frac1, match_frac2, compare_seq = contam.match(
                        seq, seqrc)
                    if match_frac1 < best_match_frac[0]:
                        continue
                    if (
                            contam.seq in compare_seq or
                            align(
                                compare_seq, contam.seq,
                                self.min_contaminant_match_frac)):
                        if (match_frac1 > best_match_frac[0] or (
                                match_frac1 == best_match_frac[0] and
                                match_frac2 > best_match_frac[1])):
                            best_matches = {}
                            best_match_frac = (match_frac1, match_frac2)
                        best_matches[contam] = (
                            match, (match_frac1, match_frac2))
                return (best_matches, best_match_frac)
            
            for match in matches:
                best_matches, best_match_frac = find_best_match(
                    match.seq, {}, (self.min_contaminant_match_frac, 0))
                
                if match.longest_match:
                    best_matches, best_match_frac = find_best_match(
                        match.longest_match[0], best_matches, best_match_frac)
                
                if best_matches:
                    for contam, match in best_matches.items():
                        if contam not in known or match[1] > known[contam][1]:
                            known[contam] = match
                else:
                    unknown.append(match)
            
            # resolve many-many relationships
            
            new_matches = defaultdict(lambda: [])
            for contam, (match, match_frac) in known.items():
                new_matches[match].append((contam, match_frac))
            
            known = []
            for match, contams in new_matches.items():
                if len(contams) == 1:
                    contam, match_frac = contams[0]
                    match.set_contaminant(contam, *match_frac)
                else:
                    contams.sort(key=lambda x: x[1], reverse=True)
                    contam, match_frac = contams[0]
                    equiv = [
                        other_contam
                        for other_contam in contams[1:]
                        if other_contam[1] == match_frac]
                    if len(equiv) == 0:
                        match.set_contaminant(contam, *match_frac)
                    else:
                        names = set(contam.names)
                        seqs = set((contam.seq,))
                        for other_contam in equiv:
                            names.update(other_contam[0].names)
                            seqs.add(other_contam[0].seq)
                        match.set_known(list(names), list(seqs), *match_frac)
                known.append(match)
            
            matches = known + unknown
        
        return matches

class KhmerDetector(Detector):
    """Identify contaminants based on kmer frequency using a fast kmer counting
    approach (as implemented in the khmer library). This approach is fast but
    not as accurate as the other two.
    """
    @property
    def min_report_freq(self):
        return 0.0001
    
    def _get_contaminants(self):
        from khmer import Countgraph, khmer_args
        # assuming all sequences are same length
        n_win = self._read_length - self.kmer_size + 1
        tablesize = self.n_reads * n_win
        countgraph = Countgraph(
            self.kmer_size, tablesize, khmer_args.DEFAULT_N_TABLES)
        countgraph.set_use_bigcount(True)
        
        for seq in self._read_sequences:
            countgraph.consume_and_tag(seq)
        
        n_expected = math.ceil(tablesize / float(4**self.kmer_size))
        min_count = n_expected * self.overrep_cutoff
        if min_count >= 2**16:
            raise ValueError(
                "The minimum count for an over-represented k-kmer {} is "
                "greater than the max khmer count (2^16)".format(min_count))
    
        candidates = {}
        
        for tag in countgraph.get_tagset():
            count = countgraph.get(tag)
            if count >= min_count:
                candidates[tag] = count
        
        if self.known_contaminants:
            matches = []
            seen = set()
            
            def match(kmer):
                """Returns the frequency of `kmer` in `candidates`.
                """
                freq = candidates.get(kmer, 0)
                if freq > 0:
                    seen.add(kmer)
                return freq
            
            for seq, names in self.known_contaminants.iter_sequences():
                seqlen = len(seq)
                if seqlen < self.kmer_size:
                    print("Cannot check {}; sequence is shorter than {}".format(
                        list(names)[0], self.kmer_size))
                    continue
                
                n_kmers = seqlen - self.kmer_size + 1
                num_matches = 0
                match_counts = []
                for idx in range(n_kmers):
                    kmer = seq[idx:(idx + self.kmer_size)]
                    kmer_count = max(
                        match(kmer),
                        match(reverse_complement(kmer))
                    )
                    if kmer_count > 0:
                        num_matches += 1
                        match_counts.append(kmer_count)
                
                if num_matches > 0:
                    # not sure what the correct metric is to use here
                    overall_count = sum(match_counts) / float(n_kmers)
                    matches.append(Match(
                        seq, overall_count / float(tablesize), names,
                        float(num_matches) / n_kmers))
            
            # Add remaining tags
            for tag in set(candidates.keys()) - seen:
                matches.append(Match(tag, candidates[tag] / float(tablesize)))
        
        else:
            matches = [
                Match(tag, count / float(tablesize))
                for tag, count in candidates.items()]
        
        return matches

def align(seq1, seq2, min_overlap_frac=0.9):
    """Align two sequences.
    
    Args:
        seq1, seq2: The sequences to align.
        min_overlap_frac: Minimum fraction of overlapping bases required for a
            match.
    
    Returns:
        The matching portion of the sequence.
    """
    aligner = Aligner(
        seq1, 0.0,
        SEMIGLOBAL,
        False, False)
    aligner.min_overlap = math.ceil(
        min(len(seq1), len(seq2)) * min_overlap_frac)
    aligner.indel_cost = 100000
    match = aligner.locate(seq2)
    if match:
        return seq1[match[0]:match[1]]
    else:
        return None

# TODO: move report to the report package, have summarize update the
# summary dict.

def summarize_contaminants(outstream, matches, n_reads, header=None):
    """Prints a report.
    """
    print("", file=outstream)
    
    if header:
        print(header, file=outstream)
        print('-' * len(header), file=outstream)
    
    n_matches = len(matches)
    print(
        "Detected {} adapters/contaminants:".format(n_matches),
        file=outstream)
    
    if n_matches == 0:
        print("Try increasing --max-reads", file=outstream)
        return
    
    pad = len(str(len(matches)))
    pad2 = ' ' * (pad + 2)
    def println(string):
        """Print a string with padding.
        """
        print(pad2 + string, file=outstream)
    
    for idx, match in enumerate(matches):
        print(
            ("{:>" + str(pad) + "}. Longest kmer: {}").format(idx+1, match.seq),
            file=outstream)
        if match.longest_match:
            println("Longest matching sequence: {}".format(
                match.longest_match[0]))
        if match.is_known:
            println("Name(s): {}".format(
                ",\n{}".format(' ' * (pad + 11)).join(match.names)))
            println("Known sequence(s): {}".format(
                ",\n{}".format(' ' * (pad + 11)).join(match.known_seqs)))
            println(
                "Known sequence K-mers that match detected contaminant: "
                "{:.2%}".format(match.match_frac))
        if match.abundance:
            println("Abundance (full-length) in {} reads: {} ({:.1%})".format(
                n_reads, match.abundance, match.abundance / n_reads))
        if match.match_frac2:
            println(
                "Detected contaminant kmers that match known sequence: "
                "{:.2%}".format(match.match_frac2))
        if match.count_is_frequency:
            println("Frequency of k-mers: {:.2%}".format(match.count))
        else:
            println("Number of k-mer matches: {}".format(match.count))



"""Implementation of the 'trim' command.
"""
from collections import Sequence, defaultdict
import logging
import os
import sys
import textwrap
from atropos.commands.base import (
    BaseCommandRunner, Summary, Pipeline, SingleEndPipelineMixin,
    PairedEndPipelineMixin)
from atropos.commands.stats import (
    SingleEndReadStatistics, PairedEndReadStatistics)
from atropos.adapters import AdapterParser, BACK
from atropos.io import STDOUT
from atropos.util import RandomMatchProbability, Const, run_interruptible
from .modifiers import (
    AdapterCutter, DoubleEncoder, InsertAdapterCutter, LengthTagModifier,
    MergeOverlapping, MinCutter, NEndTrimmer, NextseqQualityTrimmer,
    NonDirectionalBisulfiteTrimmer, OverwriteRead, PairedEndModifiers,
    PrefixSuffixAdder, PrimerTrimmer, QualityTrimmer, RRBSTrimmer,
    SingleEndModifiers, SuffixRemover, SwiftBisulfiteTrimmer,
    UnconditionalCutter, ZeroCapper)
from .filters import (
    FilterFactory, Filters, MergedReadFilter, NContentFilter, NoFilter,
    TooLongReadFilter, TooShortReadFilter, TrimmedFilter, UntrimmedFilter)
from .writers import (
    Formatters, InfoFormatter, RestFormatter, WildcardFormatter, Writers)

class TrimPipeline(Pipeline):
    """Base trimming pipeline.
    
    Args:
        record_handler:
        result_handler:
    """
    def __init__(self, record_handler, result_handler):
        super().__init__()
        self.record_handler = record_handler
        self.result_handler = result_handler
    
    def start(self, worker=None):
        self.result_handler.start(worker)
    
    def add_to_context(self, context):
        context['results'] = defaultdict(lambda: [])
    
    def handle_records(self, context, records):
        super().handle_records(context, records)
        self.result_handler.write_result(context['index'], context['results'])
    
    def handle_reads(self, context, read1, read2=None):
        return self.record_handler.handle_record(context, read1, read2)
    
    def finish(self, summary, **kwargs):
        self.result_handler.finish()
        super().finish(summary)
        summary.update(self.record_handler.summarize())

class RecordHandler(object):
    """Base class for record handlers.
    """
    def __init__(self, modifiers, filters, formatters):
        self.modifiers = modifiers
        self.filters = filters
        self.formatters = formatters
    
    def handle_record(self, context, read1, read2=None):
        """Handle a pair of reads.
        """
        reads = self.modifiers.modify(read1, read2)
        dest = self.filters.filter(*reads)
        self.formatters.format(context['results'], dest, *reads)
        return (dest, reads)
    
    def summarize(self):
        """Returns a summary dict.
        """
        return dict(trim=dict(
            modifiers=self.modifiers.summarize(),
            filters=self.filters.summarize(),
            formatters=self.formatters.summarize()))

class StatsRecordHandlerWrapper(object):
    """Wrapper around a record handler that collects read statistics
    before and/or after trimming.
    
    Args:
        record_handler:
        paired: Whether reads are paired-end.
        mode: Collection mode; pre=only collect pre-trim stats; post=only
            collect post-trim stats; both=collect both pre- and post-trim stats.
    """
    def __init__(self, record_handler, paired, stats_args, **kwargs):
        self.record_handler = record_handler
        self.read_statistics_class = (
            PairedEndReadStatistics if paired else SingleEndReadStatistics)
        self.pre = self.post = None
        if 'pre' in stats_args:
            self.pre = {}
            self.pre_kwargs = kwargs.copy()
            self.pre_kwargs.update(stats_args['pre'])
        if 'post' in stats_args:
            self.post = {}
            self.post_kwargs = kwargs.copy()
            self.post_kwargs.update(stats_args['post'])
    
    def handle_record(self, context, read1, read2=None):
        """Handle a pair of reads.
        """
        if self.pre is not None:
            self.collect(
                self.pre, context['source'], read1, read2, **self.pre_kwargs)
        dest, reads = self.record_handler.handle_record(context, read1, read2)
        if self.post is not None:
            if dest not in self.post:
                self.post[dest] = {}
            self.collect(
                self.post[dest], context['source'], *reads, **self.post_kwargs)
        return (dest, reads)
    
    def collect(self, stats, source, read1, read2=None, **kwargs):
        """Collect stats on a pair of reads.
        
        Args:
            stats: The :class:`ReadStatistics` object.
            source: The source file(s).
            read1, read2: The reads.
        """
        if source not in stats:
            stats[source] = self.read_statistics_class(**kwargs)
        stats[source].collect(read1, read2)
    
    def summarize(self):
        """Returns a summary dict.
        """
        summary = self.record_handler.summarize()
        if self.pre is not None:
            summary['pre'] = dict(
                (source, stats.summarize())
                for source, stats in self.pre.items())
        if self.post is not None:
            summary['post'] = {}
            for dest, stats_dict in self.post.items():
                summary['post'][dest.name] = dict(
                    (source, stats.summarize())
                    for source, stats in stats_dict.items())
        return summary

class ResultHandler(object):
    """Base class for result handlers.
    """
    def start(self, worker):
        """Start the result handler.
        """
        pass
    
    def finish(self, total_batches=None):
        """Finish the result handler.
        
        Args:
            total_batches: Total number of batches processed.
        """
        pass
    
    def write_result(self, batch_num, result):
        """Write a batch of results to output.
        
        Args:
            batch_num: The batch number.
            result: The result to write.
        """
        raise NotImplementedError()

class ResultHandlerWrapper(ResultHandler):
    """Wraps a ResultHandler.
    """
    def __init__(self, handler):
        self.handler = handler
    
    def start(self, worker):
        self.handler.start(worker)
    
    def write_result(self, batch_num, result):
        self.handler.write_result(batch_num, result)
    
    def finish(self, total_batches=None):
        self.handler.finish(total_batches=total_batches)

class WorkerResultHandler(ResultHandlerWrapper):
    """Wraps a ResultHandler and compresses results prior to writing.
    """
    def write_result(self, batch_num, result):
        """Given a dict mapping files to lists of strings, join the strings and
        compress them (if necessary) and then return the property formatted
        result dict.
        """
        self.handler.write_result(
            batch_num, dict(
                self.prepare_file(*item)
                for item in result.items()))
    
    def prepare_file(self, path, strings):
        """Prepare data for writing.
        
        Returns:
            Tuple (path, data).
        """
        return (path, "".join(strings))

class WriterResultHandler(ResultHandler):
    """ResultHandler that writes results to disk.
    
    Args:
        writers: :class:`Writers` object.
        compressed: Whether the data is compressed.
        use_suffix: Whether to add the worker index as a file suffix. Used for
            parallel-write mode.
    """
    def __init__(self, writers, compressed=False, use_suffix=False):
        self.writers = writers
        self.compressed = compressed
        self.use_suffix = use_suffix
    
    def start(self, worker):
        if self.use_suffix:
            self.writers.suffix = ".{}".format(worker.index)
    
    def write_result(self, batch_num, result):
        self.writers.write_result(result, self.compressed)
    
    def finish(self, total_batches=None):
        self.writers.close()

class TrimSummary(Summary):
    """Summary that adds aggregate values for record and bp stats.
    """
    def _post_process_other(self, dict_val, key, value):
        """For trim stats, any value with a name that starts with 'records_'
        will have 'fraction_<var>' computed as value / total_records, and any
        value with a name that starts with 'bp_' will have 'fraction_<var>' and
        'total_<var>' computed. We also replace any `Const`s with their values.
        """
        if self.has_exception:
            return
        
        def frac(val, total):
            """Compute fraction of total.
            """
            return (val / total) if val and total != 0 else 0
        
        if isinstance(key, str):
            if key.startswith('records_'):
                frac_key = "fraction_{}".format(key)
                total_records = self['total_record_count']
                if isinstance(value, Sequence):
                    dict_val[frac_key] = [
                        frac(val, total_records) for val in value]
                    total = sum(val for val in value if val)
                    dict_val["total_{}".format(key)] = total
                else:
                    dict_val[frac_key] = frac(value, total_records)
            elif key.startswith('bp_'):
                frac_key = "fraction_{}".format(key)
                sum_total_bp = self['sum_total_bp_count']
                if isinstance(value, Sequence):
                    dict_val[frac_key] = [
                        frac(val, bps)
                        for val, bps in zip(value, self['total_bp_counts'])]
                    total = sum(val for val in value if val)
                    dict_val["total_{}".format(key)] = total
                    dict_val["fraction_total_{}".format(key)] = \
                        frac(total, sum_total_bp)
                else:
                    dict_val[frac_key] = frac(value, sum_total_bp)

class CommandRunner(BaseCommandRunner):
    name = 'trim'
    
    def __init__(self, options):
        super().__init__(options, TrimSummary)
    
    def __call__(self):
        options = self.options
        match_probability = RandomMatchProbability()
        
        # Create Adapters
        
        has_adapters1 = options.adapters or options.anywhere or options.front
        has_adapters2 = options.adapters2 or options.anywhere2 or options.front2
        
        adapters1 = adapters2 = []
        if has_adapters1 or has_adapters2:
            adapter_cache = super().load_known_adapters()
            parser_args = dict(
                colorspace=options.colorspace,
                max_error_rate=options.error_rate,
                min_overlap=options.overlap,
                read_wildcards=options.match_read_wildcards,
                adapter_wildcards=options.match_adapter_wildcards,
                indels=options.indels, indel_cost=options.indel_cost,
                cache=adapter_cache, gc_content=options.gc_content,
                match_probability=match_probability)
            if options.adapter_max_rmp:
                parser_args['max_rmp'] = options.adapter_max_rmp
            adapter_parser = AdapterParser(**parser_args)
            
            if has_adapters1:
                adapters1 = adapter_parser.parse_multi(
                    options.adapters, options.anywhere, options.front)
            if has_adapters2:
                adapters2 = adapter_parser.parse_multi(
                    options.adapters2, options.anywhere2, options.front2)
            
            if options.cache_adapters:
                adapter_cache.save()
        
        # Create Modifiers
        
        # TODO: can this be replaced with an argparse required group?
        if (
                not adapters1 and not adapters2 and
                not options.quality_cutoff and
                options.nextseq_trim is None and
                options.cut == [] and options.cut2 == [] and
                options.cut_min == [] and options.cut_min2 == [] and
                (
                    options.minimum_length is None or
                    options.minimum_length <= 0) and
                options.maximum_length == sys.maxsize and not options.trim_n and
                not self.has_qualfile and options.max_n is None and
                (not options.paired or options.overwrite_low_quality is None)):
            raise ValueError(
                "You need to provide at least one adapter sequence.")
        
        if (
                options.aligner == 'insert' and any(
                    not a or len(a) != 1 or a[0].where != BACK
                    for a in (adapters1, adapters2))):
            raise ValueError(
                "Insert aligner requires a single 3' adapter for each read")
        
        if options.debug:
            for adapter in adapters1 + adapters2:
                adapter.enable_debug()
        
        if options.paired:
            modifiers = PairedEndModifiers(options.paired)
        else:
            modifiers = SingleEndModifiers()
        
        for oper in options.op_order:
            if oper == 'W' and options.overwrite_low_quality:
                lowq, highq, window = options.overwrite_low_quality
                modifiers.add_modifier(
                    OverwriteRead,
                    worse_read_min_quality=lowq, better_read_min_quality=highq,
                    window_size=window, base=options.quality_base)
                
            elif oper == 'A' and (adapters1 or adapters2):
                # TODO: generalize this using some kind of factory class
                if options.aligner == 'insert':
                    # Use different base probabilities if we're trimming
                    # bisulfite data.
                    # TODO: this doesn't seem to help things, so commenting it
                    # out for now
                    #if options.bisulfite:
                    #   base_probs = dict(match_prob=0.33, mismatch_prob=0.67)
                    # else:
                    #   base_probs = dict(match_prob=0.25, mismatch_prob=0.75)
                    modifiers.add_modifier(
                        InsertAdapterCutter,
                        adapter1=adapters1[0], adapter2=adapters2[0],
                        action=options.action,
                        mismatch_action=options.correct_mismatches,
                        max_insert_mismatch_frac=\
                            options.insert_match_error_rate,
                        max_adapter_mismatch_frac=\
                            options.insert_match_adapter_error_rate,
                        match_probability=match_probability,
                        insert_max_rmp=options.insert_max_rmp)
                else:
                    a1_args = dict(
                        adapters=adapters1,
                        times=options.times,
                        action=options.action) if adapters1 else None
                    a2_args = dict(
                        adapters=adapters2,
                        times=options.times,
                        action=options.action) if adapters2 else None
                    modifiers.add_modifier_pair(AdapterCutter, a1_args, a2_args)
            elif oper == 'C' and (options.cut or options.cut2):
                modifiers.add_modifier_pair(
                    UnconditionalCutter,
                    dict(lengths=options.cut),
                    dict(lengths=options.cut2))
            elif oper == 'G' and (options.nextseq_trim is not None):
                modifiers.add_modifier(
                    NextseqQualityTrimmer,
                    read=1, cutoff=options.nextseq_trim,
                    base=options.quality_base)
            elif oper == 'Q' and options.quality_cutoff:
                modifiers.add_modifier(
                    QualityTrimmer,
                    cutoff_front=options.quality_cutoff[0],
                    cutoff_back=options.quality_cutoff[1],
                    base=options.quality_base)
        
        if options.bisulfite:
            if isinstance(options.bisulfite, str):
                if "non-directional" in options.bisulfite:
                    modifiers.add_modifier(
                        NonDirectionalBisulfiteTrimmer,
                        rrbs=options.bisulfite=="non-directional-rrbs")
                elif options.bisulfite == "rrbs":
                    modifiers.add_modifier(RRBSTrimmer)
                elif options.bisulfite in ("epignome", "truseq"):
                    # Trimming leads to worse results
                    #modifiers.add_modifier(TruSeqBisulfiteTrimmer)
                    pass
                elif options.bisulfite == "swift":
                    modifiers.add_modifier(SwiftBisulfiteTrimmer)
            else:
                if options.bisulfite[0]:
                    modifiers.add_modifier(
                        MinCutter, read=1, **(options.bisulfite[0]))
                if len(options.bisulfite) > 1 and options.bisulfite[1]:
                    modifiers.add_modifier(
                        MinCutter, read=2, **(options.bisulfite[1]))
        
        if options.trim_n:
            modifiers.add_modifier(NEndTrimmer)
        
        if options.cut_min or options.cut_min2:
            modifiers.add_modifier_pair(
                MinCutter,
                dict(lengths=options.cut_min),
                dict(lengths=options.cut_min2))
        
        if options.length_tag:
            modifiers.add_modifier(
                LengthTagModifier, length_tag=options.length_tag)
        
        if options.strip_suffix:
            modifiers.add_modifier(SuffixRemover, suffixes=options.strip_suffix)
        
        if options.prefix or options.suffix:
            modifiers.add_modifier(
                PrefixSuffixAdder, prefix=options.prefix, suffix=options.suffix)
        
        if options.double_encode:
            modifiers.add_modifier(DoubleEncoder)
        
        if options.zero_cap and self.delivers_qualities:
            modifiers.add_modifier(
                ZeroCapper, quality_base=options.quality_base)
        
        if options.trim_primer:
            modifiers.add_modifier(PrimerTrimmer)
        
        if options.merge_overlapping:
            modifiers.add_modifier(
                MergeOverlapping,
                min_overlap=options.merge_min_overlap,
                error_rate=options.merge_error_rate,
                mismatch_action=options.correct_mismatches)
        
        # Create Filters and Formatters
        
        min_affected = 2 if options.pair_filter == 'both' else 1
        filters = Filters(FilterFactory(options.paired, min_affected))
        
        output1 = output2 = None
        interleaved = False
        if options.interleaved_output:
            output1 = options.interleaved_output
            interleaved = True
        else:
            output1 = options.output
            output2 = options.paired_output
        
        seq_formatter_args = dict(
            qualities=self.delivers_qualities,
            colorspace=options.colorspace,
            interleaved=interleaved
        )
        formatters = Formatters(output1, seq_formatter_args)
        force_create = []
            
        if options.merge_overlapping:
            filters.add_filter(MergedReadFilter)
            if options.merged_output:
                formatters.add_seq_formatter(
                    MergedReadFilter, options.merged_output)
            
        if options.minimum_length is not None and options.minimum_length > 0:
            filters.add_filter(TooShortReadFilter, options.minimum_length)
            if options.too_short_output:
                formatters.add_seq_formatter(
                    TooShortReadFilter,
                    options.too_short_output, options.too_short_paired_output)
        
        if options.maximum_length < sys.maxsize:
            filters.add_filter(TooLongReadFilter, options.maximum_length)
            if options.too_long_output is not None:
                formatters.add_seq_formatter(
                    TooLongReadFilter,
                    options.too_long_output, options.too_long_paired_output)
        
        if options.max_n is not None:
            filters.add_filter(NContentFilter, options.max_n)
        
        if options.discard_trimmed:
            filters.add_filter(TrimmedFilter)
        
        if not formatters.multiplexed:
            if output1 is not None:
                formatters.add_seq_formatter(NoFilter, output1, output2)
                if output1 != STDOUT and options.writer_process:
                    force_create.append(output1)
                    if output2 is not None:
                        force_create.append(output2)
            elif not (options.discard_trimmed and options.untrimmed_output):
                formatters.add_seq_formatter(NoFilter, options.default_outfile)
                if options.default_outfile != STDOUT and options.writer_process:
                    force_create.append(options.default_outfile)
        
        if options.discard_untrimmed or options.untrimmed_output:
            filters.add_filter(UntrimmedFilter)
        
        if not options.discard_untrimmed:
            if formatters.multiplexed:
                untrimmed = options.untrimmed_output or output1.format(
                    name='unknown')
                formatters.add_seq_formatter(UntrimmedFilter, untrimmed)
                formatters.add_seq_formatter(NoFilter, untrimmed)
            elif options.untrimmed_output:
                formatters.add_seq_formatter(
                    UntrimmedFilter,
                    options.untrimmed_output, options.untrimmed_paired_output)
        
        if options.rest_file:
            formatters.add_info_formatter(RestFormatter(options.rest_file))
        if options.info_file:
            formatters.add_info_formatter(InfoFormatter(options.info_file))
        if options.wildcard_file:
            formatters.add_info_formatter(
                WildcardFormatter(options.wildcard_file))
        
        if options.paired:
            mixin_class = PairedEndPipelineMixin
        else:
            mixin_class = SingleEndPipelineMixin
        writers = Writers(force_create)
        record_handler = RecordHandler(modifiers, filters, formatters)
        if options.stats:
            record_handler = StatsRecordHandlerWrapper(
                record_handler, options.paired, options.stats,
                qualities=self.delivers_qualities,
                quality_base=self.quality_base)
        
        logger = logging.getLogger()
        num_adapters = sum(len(a) for a in modifiers.get_adapters())
        logger.info(
            "Trimming %s adapter%s with at most %.1f%% errors in %s mode ...",
            num_adapters,
            's' if num_adapters > 1 else '', options.error_rate * 100,
            {
                False: 'single-end',
                'first': 'paired-end legacy',
                'both': 'paired-end'
            }[options.paired])
        if (
                options.paired == 'first' and (
                    len(record_handler.modifiers.get_modifiers(read=2)) > 0 or
                    options.quality_cutoff)):
            logger.warning('\n'.join(textwrap.wrap(
                'WARNING: Requested read modifications are applied only to the '
                'first read since backwards compatibility mode is enabled. '
                'To modify both reads, also use any of the -A/-B/-G/-U '
                'options. Use a dummy adapter sequence when necessary: '
                '-A XXX')))
        
        if options.threads is None:
            # Run single-threaded version
            result_handler = WorkerResultHandler(WriterResultHandler(writers))
            pipeline_class = type(
                'TrimPipelineImpl', (mixin_class, TrimPipeline), {})
            pipeline = pipeline_class(record_handler, result_handler)
            self.summary.update(mode='serial', threads=1)
            return run_interruptible(pipeline, self, raise_on_error=True)
        else:
            # Run multiprocessing version
            self.summary.update(mode='parallel', threads=options.threads)
            return self.run_parallel(record_handler, writers, mixin_class)
    
    def run_parallel(self, record_handler, writers, mixin_class):
        """Parallel implementation of run_atropos. Works as follows:
        
        1. Main thread creates N worker processes (where N is the number of
        threads to be allocated) and (optionally) one writer process.
        2. Main thread loads batches of reads (or read pairs) from input file(s)
        and adds them to a queue (the input queue).
        3. Worker processes take batches from the input queue, process them as
        Atropos normally does, and either add the results to the result queue
        (if using a writer process) or write the results to disk. Each result is
        a dict mapping output file names to strings, where each string is a
        concatenation of reads (with appropriate line endings) to be written.
        A parameter also controls whether data compression is done by the
        workers or the writer.
        4. If using a writer process, it takes results from the result queue and
        writes each string to its corresponding file.
        5. When the main process finishes loading reads from the input file(s),
        it sends a signal to the worker processes that they should complete when
        the input queue is empty. It also singals the writer process how many
        total batches to expect, and the writer process exits after it has
        processed that many batches.
        6. When a worker process completes, it adds a summary of its activity to
        the summary queue.
        7. The main process reads summaries from the summary queue and merges
        them to create the complete summary, which is used to generate the
        report.
        
        There are several possible points of failure:
        
        1. The main process may exit due to an unexpected error, or becuase the
        user forces it to exit (Ctrl-C). In this case, an attempt is made to
        cancel all processes before exiting.
        2. A worker or writer process may exit due to an unknown error. To
        handle this, the main process checks that each process is alive whenver
        it times out writing to the input queue, and again when waiting for
        worker summaries. If a process has died, the program exits with an error
        since some data might have gotten lost.
        3. More commonly, process will time out blocking on reading from or
        writing to a queue. Size limits are used (optionally) for the input and
        result queues to prevent using lots of memory. When few threads are
        allocated, it is most likely that the main and writer processes will
        block, whereas with many threads allocated the workers are most likely
        to block. Also, e.g. in a cluster environment, I/O latency may cause a
        "backup" resulting in frequent blocking of the main and workers
        processes. Finally, also e.g. in a cluster environment, processes may
        suspended for periods of time. Use of a hard timeout period, after which
        processes are forced to exit, is thus undesirable. Instead, parameters
        are provided for the user to tune the batch size and max queue sizes to
        their particular environment. Additionally, a "soft" timeout is used,
        after which log messages are escallated from DEBUG to ERROR level. The
        user can then make the decision of whether or not to kill the program.
        
        Args:
            record_handler: RecordHandler object.
            writers: Writers object.
            mixin_class: Mixin to use for creating pipeline class.
        
        Returns:
            The return code.
        """
        # We do all the multicore imports and class definitions within the
        # run_parallel method to avoid extra work if only running in serial
        # mode.
        from multiprocessing import Process, Queue
        from atropos.commands.multicore import (
            Control, PendingQueue, ParallelPipelineMixin,
            ParallelPipelineRunner, MulticoreError, wait_on_process, enqueue,
            dequeue, kill, RETRY_INTERVAL, CONTROL_ACTIVE, CONTROL_ERROR)
        from atropos.io.compression import (
            get_compressor, can_use_system_compression)
        
        class Done(MulticoreError):
            """Raised when process exits normally.
            """
            pass
        
        class Killed(MulticoreError):
            """Raised when process exits is killed.
            """
            pass
        
        class ParallelTrimPipelineRunner(ParallelPipelineRunner):
            """ParallelPipelineRunner for a TrimPipeline.
            """
            def __init__(
                    self, command_runner, pipeline, threads, writer_manager=None):
                super().__init__(command_runner, pipeline, threads)
                self.writer_manager = writer_manager
            
            def ensure_alive(self):
                super().ensure_alive()
                if self.writer_manager and not self.writer_manager.is_active():
                    raise MulticoreError("Writer process exited")
            
            def after_enqueue(self):
                # Tell the writer thread the max number of batches to expect
                if self.writer_manager:
                    self.writer_manager.set_num_batches(self.num_batches)
            
            def finish(self):
                if self.writer_manager:
                    # Wait for writer to complete
                    self.writer_manager.wait()
            
            def terminate(self, retcode):
                super().terminate(retcode)
                if self.writer_manager:
                    self.writer_manager.terminate(retcode)
        
        class QueueResultHandler(ResultHandler):
            """ResultHandler that writes results to the output queue.
            """
            def __init__(self, queue):
                self.queue = queue
                self.message = None
                self.timeout = None
            
            def start(self, worker):
                self.message = "{} waiting to queue result {{}}".format(
                    worker.name)
                self.timeout = worker.timeout
            
            def write_result(self, batch_num, result):
                enqueue(
                    self.queue,
                    (batch_num, result),
                    wait_message=self.message,
                    timeout=self.timeout)
        
        class CompressingWorkerResultHandler(WorkerResultHandler):
            """Wraps a ResultHandler and compresses results prior to writing.
            """
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.file_compressors = None
            
            def start(self, worker):
                super().start(worker)
                self.file_compressors = {}
            
            def prepare_file(self, path, strings):
                compressor = self.get_compressor(path)
                if compressor:
                    return ((path, 'wb'), compressor.compress(b''.join(
                        s.encode() for s in strings)))
                else:
                    return ((path, 'wt'), "".join(strings))
            
            def get_compressor(self, filename):
                """Returns the file compressor based on the file extension.
                """
                if filename not in self.file_compressors:
                    self.file_compressors[filename] = get_compressor(filename)
                return self.file_compressors[filename]
        
        class OrderPreservingWriterResultHandler(WriterResultHandler):
            """Writer thread that is less time/memory efficient, but is
            guaranteed to preserve the original order of records.
            """
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.pending = None
                self.cur_batch = None
            
            def start(self, worker):
                super().__init__(worker)
                self.pending = PendingQueue()
                self.cur_batch = 1
            
            def write_result(self, batch_num, result):
                if batch_num == self.cur_batch:
                    self.writers.write_result(result, self.compressed)
                    self.cur_batch += 1
                    self.consume_pending()
                else:
                    self.pending.push(batch_num, result)
            
            def finish(self, total_batches=None):
                if total_batches is not None:
                    self.consume_pending()
                    if self.cur_batch != total_batches:
                        raise MulticoreError(
                            "OrderPreservingWriterResultHandler finishing "
                            "without having seen {} batches".format(
                                total_batches))
                self.writers.close()
            
            def consume_pending(self):
                """Consume any remaining items in the queue.
                """
                while (
                        (not self.pending.empty) and
                        (self.cur_batch == self.pending.min_priority)):
                    self.writers.write_result(
                        self.pending.pop(), self.compressed)
                    self.cur_batch += 1
        
        class ResultProcess(Process):
            """Thread that accepts results from the worker threads and process
            them using a ResultHandler. Each batch is expected to be
            (batch_num, path, records), where path is the destination file and
            records is a string. Not guaranteed to preserve the original order
            of sequence records.
            
            Args:
                result_handler: A ResultHandler object.
                queue: Input queue.
                control: A shared value for communcation with the main process.
                timeout: Seconds to wait for next batch before complaining.
            """
            def __init__(self, result_handler, queue, control, timeout=60):
                super().__init__(name="Result process")
                self.result_handler = result_handler
                self.queue = queue
                self.control = control
                self.timeout = timeout
                self.seen_batches = set()
                self.num_batches = None
            
            def run(self):
                logging.getLogger().debug(
                    "Writer process %s running under pid %d",
                    self.name, os.getpid())
                
                def fail_callback():
                    """Raises Done if the expected number of batches has been
                    seen.
                    """
                    if (
                            self.num_batches is None and
                            self.control.check_value_positive()):
                        self.num_batches = self.control.get_value()
                    if (
                            self.num_batches is not None and
                            len(self.seen_batches) >= self.num_batches):
                        raise Done()
                
                def timeout_callback():
                    """Logs an error with the missing batches.
                    """
                    if self.num_batches is not None:
                        missing = (
                            set(range(1, self.num_batches+1)) -
                            self.seen_batches)
                        logging.getLogger().error(
                            "Result thread still missing batches %s of %d",
                            ",".join(str(i) for i in missing),
                            self.num_batches)
                
                def iter_batches():
                    """Dequeue and yield batches.
                    """
                    while True:
                        batch = dequeue(
                            self.queue,
                            wait_message="Result process waiting on result {}",
                            timeout=self.timeout,
                            fail_callback=fail_callback,
                            timeout_callback=timeout_callback)
                        yield batch
                
                try:
                    self.result_handler.start(self)
                    
                    for batch_num, result in iter_batches():
                        self.seen_batches.add(batch_num)
                        self.result_handler.write_result(batch_num, result)
                except Done:
                    logging.getLogger().debug("Writer process exiting normally")
                except Killed:
                    logging.getLogger().debug("Writer process exited early")
                except:
                    logging.getLogger().error(
                        "Unexpected error in writer process", exc_info=True)
                    self.control.set_value(CONTROL_ERROR)
                finally:
                    num_batches = self.control.get_value(lock=True)
                    self.result_handler.finish(
                        num_batches if num_batches > 0 else None)
        
        class WriterManager(object):
            """Manager for a writer process and control variable.
            """
            def __init__(
                    self, writers, compression, preserve_order, result_queue,
                    timeout):
                # result handler
                if preserve_order:
                    writer_result_handler = OrderPreservingWriterResultHandler(
                        writers, compressed=compression == "worker")
                else:
                    writer_result_handler = WriterResultHandler(
                        writers, compressed=compression == "worker")
                
                self.timeout = timeout
                # Shared variable for communicating with writer thread
                self.writer_control = Control(CONTROL_ACTIVE)
                # writer process
                self.writer_process = ResultProcess(
                    writer_result_handler, result_queue, self.writer_control,
                    timeout)
                self.writer_process.start()
            
            def is_active(self):
                """Returns True if the writer process is alive and the control
                value is CONTROL_ACTIVE.
                """
                return (
                    self.writer_process.is_alive() and
                    self.writer_control.check_value(CONTROL_ACTIVE))
            
            def set_num_batches(self, num_batches):
                """Set the number of batches to the control variable.
                """
                self.writer_control.set_value(num_batches)
            
            def wait(self):
                """Wait for the writer process to terminate.
                """
                wait_on_process(self.writer_process, self.timeout)
            
            def terminate(self, retcode):
                """Force the writer process to terminate.
                """
                kill(self.writer_process, retcode, self.timeout)
        
        # Main process
        
        timeout = max(self.process_timeout, RETRY_INTERVAL)
        threads = self.threads
        
        logging.getLogger().debug(
            "Starting atropos in parallel mode with threads=%d, timeout=%d",
            threads, timeout)
        
        if threads < 2:
            raise ValueError("'threads' must be >= 2")
        
        # Reserve a thread for the writer process if it will be doing the
        # compression and if one is available.
        compression = self.compression
        if compression is None:
            compression = "worker"
            if self.writer_process and can_use_system_compression():
                compression = "writer"
        if compression == "writer" and threads > 2:
            threads -= 1
        
        # Queue by which results are sent from the worker processes to the
        # writer process
        result_queue = Queue(self.result_queue_size)
        writer_manager = None
        
        if self.writer_process:
            if compression == "writer":
                worker_result_handler = WorkerResultHandler(
                    QueueResultHandler(result_queue))
            else:
                worker_result_handler = CompressingWorkerResultHandler(
                    QueueResultHandler(result_queue))
            writer_manager = WriterManager(
                writers, compression, self.preserve_order, result_queue,
                timeout)
        else:
            worker_result_handler = WorkerResultHandler(
                WriterResultHandler(writers, use_suffix=True))
        
        pipeline_class = type(
            'TrimPipelineImpl',
            (ParallelPipelineMixin, mixin_class, TrimPipeline), {})
        pipeline = pipeline_class(record_handler, worker_result_handler)
        runner = ParallelTrimPipelineRunner(
            self, pipeline, threads, writer_manager)
        return runner.run()
